import logging
import os
import random
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Sequence, Tuple, Union

import numpy as np
import torch
from aislib.misc_utils import ensure_path_exists
from ignite.engine import Engine

from eir.train_utils.distributed import (
    in_distributed_env,
    in_master_node,
    only_call_on_master_node,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)


def get_run_folder(output_folder: str) -> Path:
    return Path(output_folder)


def prepare_sample_output_folder(
    output_folder: str, output_name: str, column_name: str, iteration: int
) -> Path:
    sample_outfolder = (
        get_run_folder(output_folder=output_folder)
        / "results"
        / output_name
        / column_name
        / "samples"
        / str(iteration)
    )
    ensure_path_exists(sample_outfolder, is_folder=True)

    return sample_outfolder


@only_call_on_master_node
def configure_global_eir_logging(output_folder: str, log_level: str = "INFO") -> None:
    level = parse_log_level(log_level=log_level)

    logfile_path = get_run_folder(output_folder=output_folder) / "logging_history.log"

    ensure_path_exists(path=logfile_path)
    file_handler = logging.FileHandler(filename=str(logfile_path))
    file_handler.setLevel(level=level)

    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(name)s - %(message)s", "%Y-%m-%d %H:%M:%S"
    )
    file_handler.setFormatter(fmt=formatter)

    root_logger = logging.getLogger(name="")
    root_logger.addHandler(hdlr=file_handler)

    set_log_level_for_eir_loggers(log_level=log_level)


def parse_log_level(log_level: str) -> int:
    log_levels = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "NOTSET": logging.NOTSET,
    }

    if log_level.upper() not in log_levels:
        raise ValueError(f"Invalid log level: {log_level}")

    return log_levels[log_level.upper()]


def set_log_level_for_eir_loggers(log_level: str = "INFO") -> None:
    level = parse_log_level(log_level=log_level)

    loggers = (logging.getLogger(name) for name in logging.root.manager.loggerDict)
    for logger_ in loggers:
        if logger_.name.split(".")[0] == "eir":
            logger_.setLevel(level=level)


def validate_handler_dependencies(handler_dependencies: Sequence[Callable]):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            argument_iterable = tuple(args) + tuple(kwargs.values())

            engine_objects = [i for i in argument_iterable if isinstance(i, Engine)]
            assert len(engine_objects) == 1

            engine_object = engine_objects[0]

            for dep in handler_dependencies:
                if not engine_object.has_event_handler(dep):
                    if in_master_node() or not in_distributed_env():
                        logger.warning(
                            f"Dependency '{dep.__name__}' missing from engine. "
                            f"If your are running EIR directly through the CLI, "
                            f"this is likely a bug. If you are customizing "
                            f"EIR (e.g. the validation handler), this can "
                            f"be expected, please ignore this warning in "
                            f"that case."
                        )

            func_output = func(*args, **kwargs)
            return func_output

        return wrapper

    return decorator


class MissingHandlerDependencyError(Exception):
    pass


def call_hooks_stage_iterable(
    hook_iterable: Iterable[Callable],
    common_kwargs: Dict,
    state: Union[None, Dict[str, Any]],
):
    for hook in hook_iterable:
        _, state = state_registered_hook_call(
            hook_func=hook, **common_kwargs, state=state
        )

    return state


def state_registered_hook_call(
    hook_func: Callable,
    state: Union[Dict[str, Any], None],
    *args,
    **kwargs,
) -> Tuple[Any, Dict[str, Any]]:
    if state is None:
        state = {}

    state_updates = hook_func(state=state, *args, **kwargs)

    state = {**state, **state_updates}

    return state_updates, state


def seed_everything(seed: int = 0) -> None:
    seed, from_os = get_seed(default_seed=seed)

    extra_log = " grabbed from environment variable 'EIR_SEED '" if from_os else " "
    logger.info("Global random seed%sset to %d.", extra_log, seed)

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def get_seed(default_seed: int = 0) -> Tuple[int, bool]:
    os_seed = os.environ.get("EIR_SEED", None)

    if os_seed:
        seed = int(os_seed)
        from_os = True
    else:
        seed = default_seed
        from_os = False

    return seed, from_os



import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Optional

import numpy as np
from aislib.misc_utils import ensure_path_exists
from sklearn.preprocessing import KBinsDiscretizer, LabelEncoder, StandardScaler

from eir.target_setup.target_setup_utils import IdentityTransformer
from eir.train_utils.utils import get_run_folder

if TYPE_CHECKING:
    from eir.data_load.label_setup import (
        al_label_transformers,
        al_label_transformers_object,
    )


def get_transformer_sources(run_folder: Path) -> dict[str, list[str]]:
    transformers_to_load: dict[str, list[str]] = {}
    transformer_sources = run_folder / "serializations/transformers"

    if not transformer_sources.exists():
        return transformers_to_load

    for transformer_source in transformer_sources.iterdir():
        names = sorted([i.stem for i in transformer_source.iterdir()])
        transformers_to_load[transformer_source.stem] = names
    return transformers_to_load


def load_transformer(
    run_folder: Path,
    source_name: str,
    transformer_name: str,
) -> "al_label_transformers_object":
    target_transformer_path = get_transformer_path(
        run_path=run_folder,
        transformer_name=transformer_name,
        source_name=source_name,
    )
    serialized_data = read_json(target_transformer_path)
    return deserialize_transformer(data=serialized_data)


def load_transformers(
    transformers_to_load: Optional[dict[str, list[str]]] = None,
    output_folder: Optional[str] = None,
    run_folder: Optional[Path] = None,
) -> dict[str, "al_label_transformers"]:
    if not run_folder and not output_folder:
        raise ValueError("Either 'run_folder' or 'output_folder' must be provided.")

    if not run_folder:
        assert output_folder is not None
        run_folder = get_run_folder(output_folder=output_folder)

    if not transformers_to_load:
        transformers_to_load = get_transformer_sources(run_folder=run_folder)

    loaded_transformers: dict[str, "al_label_transformers"] = {}
    for source_name, source_transformers_to_load in transformers_to_load.items():
        loaded_transformers[source_name] = {}
        for transformer_name in source_transformers_to_load:
            loaded_transformers[source_name][transformer_name] = load_transformer(
                run_folder=run_folder,
                source_name=source_name,
                transformer_name=transformer_name,
            )

    return loaded_transformers


def get_transformer_path(
    run_path: Path, source_name: str, transformer_name: str
) -> Path:
    if not transformer_name.endswith(".json"):
        transformer_name = f"{transformer_name}.json"

    transformer_path = (
        run_path / "serializations/transformers" / source_name / f"{transformer_name}"
    )

    return transformer_path


def save_label_transformer(
    run_folder: Path,
    output_name: str,
    transformer_name: str,
    target_transformer_object: "al_label_transformers_object",
) -> Path:
    target_transformer_outpath = get_transformer_path(
        run_path=run_folder,
        source_name=output_name,
        transformer_name=transformer_name,
    )
    ensure_path_exists(path=target_transformer_outpath)
    serialized_data = serialize_transformer(transformer=target_transformer_object)
    write_json(data=serialized_data, path=target_transformer_outpath)
    return target_transformer_outpath


def save_transformer_set(
    transformers_per_source: dict[str, "al_label_transformers"], run_folder: Path
) -> None:
    for output_name, transformers in transformers_per_source.items():
        for transformer_name, transformer_object in transformers.items():
            save_label_transformer(
                run_folder=run_folder,
                output_name=output_name,
                transformer_name=transformer_name,
                target_transformer_object=transformer_object,
            )


def serialize_standard_scaler(scaler: StandardScaler) -> dict[str, Any]:
    return {
        "type": "StandardScaler",
        "mean": scaler.mean_.tolist() if scaler.mean_ is not None else None,
        "scale": scaler.scale_.tolist() if scaler.scale_ is not None else None,
        "var": scaler.var_.tolist() if scaler.var_ is not None else None,
    }


def serialize_label_encoder(encoder: LabelEncoder) -> dict[str, Any]:
    return {
        "type": "LabelEncoder",
        "classes": encoder.classes_.tolist(),
    }


def serialize_kbins_discretizer(kbins: KBinsDiscretizer) -> dict[str, Any]:
    return {
        "type": "KBinsDiscretizer",
        "n_bins": kbins.n_bins,
        "encode": kbins.encode,
        "strategy": kbins.strategy,
        "bin_edges_": kbins.bin_edges_[0].tolist(),
    }


def serialize_identity_transformer(transformer: IdentityTransformer) -> dict[str, Any]:
    return {
        "type": "IdentityTransformer",
    }


def serialize_transformer(
    transformer: "al_label_transformers_object",
) -> dict[str, Any]:
    serializers: dict[type, Callable] = {
        StandardScaler: serialize_standard_scaler,
        LabelEncoder: serialize_label_encoder,
        KBinsDiscretizer: serialize_kbins_discretizer,
        IdentityTransformer: serialize_identity_transformer,
    }
    serializer = serializers.get(type(transformer))
    if serializer is None:
        raise ValueError(f"Unsupported transformer type: {type(transformer)}")
    return serializer(transformer)


def deserialize_standard_scaler(data: dict[str, Any]) -> StandardScaler:
    scaler = StandardScaler()
    scaler.mean_ = np.array(data["mean"]) if data["mean"] is not None else None
    scaler.scale_ = np.array(data["scale"]) if data["scale"] is not None else None
    scaler.var_ = np.array(data["var"]) if data["var"] is not None else None
    return scaler


def deserialize_label_encoder(data: dict[str, Any]) -> LabelEncoder:
    encoder = LabelEncoder()
    encoder.classes_ = np.array(data["classes"])
    return encoder


def deserialize_kbins_discretizer(data: dict[str, Any]) -> KBinsDiscretizer:
    kbins = KBinsDiscretizer(
        n_bins=data["n_bins"],
        encode=data["encode"],
        strategy=data["strategy"],
    )

    # note we need to wrap the bin_edges in an array to match the expected shape,
    # as KBinsDiscretizer expects a 2D array (if encoding multiple features, even
    # though we're only encoding one feature here)
    kbins.bin_edges_ = np.array([data["bin_edges_"]])

    # normally initialized in .fit, but we need to set it here manually for
    # various functionality to work
    kbins.n_bins_ = kbins._validate_n_bins(n_features=1)
    return kbins


def deserialize_identity_transformer(data: dict[str, Any]) -> IdentityTransformer:
    return IdentityTransformer()


def deserialize_transformer(data: dict[str, Any]) -> "al_label_transformers_object":
    deserializers: dict[str, Callable] = {
        "StandardScaler": deserialize_standard_scaler,
        "LabelEncoder": deserialize_label_encoder,
        "KBinsDiscretizer": deserialize_kbins_discretizer,
        "IdentityTransformer": deserialize_identity_transformer,
    }
    deserializer = deserializers.get(data["type"])
    if deserializer is None:
        raise ValueError(f"Unsupported transformer type: {data['type']}")
    return deserializer(data)


def write_json(data: dict[str, Any], path: Path) -> None:
    with path.open("w") as f:
        json.dump(data, f)


def read_json(path: Path) -> dict[str, Any]:
    with path.open("r") as f:
        return json.load(f)



from typing import Optional

from eir.setup.input_setup_modules.setup_sequence import get_sequence_split_function


def streamline_sequence_manual_data(
    data: str, split_on: Optional[str]
) -> list[str] | str:
    """
    This is to specifically handle the case of an empty string / None being passed
    here. If e.g. we call the split_func on '', we will get [''], which will
    end up being encoded as a <unk> token. Instead, we want to return an empty
    list here. In e.g. the validation handler code, this is also set explicitly.
    """

    sequence_streamlined: list[str] | str
    if data == "" or data is None:
        sequence_streamlined = []
    else:
        split_func = get_sequence_split_function(split_on=split_on)
        split_data = split_func(data)
        sequence_streamlined = split_data

    return sequence_streamlined



from typing import Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


class InterpolateProjection(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_size: Tuple[int, int],
        mode: str = "bilinear",
        align_corners: Optional[bool] = None,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.output_size = output_size
        self.mode = mode
        self.align_corners = align_corners

        self.sampling = ArbitrarySampling(
            output_size=output_size,
            mode=mode,
            align_corners=align_corners,
        )

        self.norm = nn.GroupNorm(num_groups=1, num_channels=self.out_channels)

        self.conv_1 = nn.Conv2d(
            in_channels=self.out_channels,
            out_channels=self.out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
        )

        self.act = nn.GELU()

        self.downsample_identity: nn.Module
        if in_channels != out_channels:
            self.downsample_identity = nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=True,
            )
        else:
            self.downsample_identity = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = self.downsample_identity(x)
        identity = self.sampling(identity)

        out = self.norm(identity)
        out = self.conv_1(out)
        out = self.act(out)

        return out + identity


class ArbitrarySampling(nn.Module):
    def __init__(
        self,
        output_size: Tuple[int, int],
        mode: str = "bilinear",
        align_corners: Optional[bool] = None,
    ):
        super().__init__()

        self.output_size = output_size
        self.mode = mode
        self.align_corners = align_corners

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(output_size={self.output_size}, "
            f"mode={self.mode}, align_corners={self.align_corners})"
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.interpolate(
            input=x,
            size=self.output_size,
            mode=self.mode,
            align_corners=self.align_corners,
        )



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, Optional, Sequence, Type, Union

from eir.models.fusion.fusion_identity import IdentityConfig
from eir.models.fusion.fusion_mgmoe import MGMoEModelConfig
from eir.models.input.array.array_models import ArrayModelConfig
from eir.models.input.image.image_models import ImageModelConfig
from eir.models.input.omics.omics_models import (
    CNNModel,
    IdentityModel,
    LCLModel,
    LinearModel,
    OmicsModelConfig,
    SimpleLCLModel,
)
from eir.models.input.sequence.transformer_models import SequenceModelConfig
from eir.models.input.tabular.tabular import SimpleTabularModel, TabularModelConfig
from eir.models.layers.mlp_layers import ResidualMLPConfig
from eir.models.output.array.array_output_modules import ArrayOutputModuleConfig
from eir.models.output.sequence.sequence_output_modules import (
    SequenceOutputModuleConfig,
)
from eir.models.output.tabular.tabular_output_modules import TabularOutputModuleConfig
from eir.setup.schema_modules.latent_analysis_schemas import LatentSamplingConfig
from eir.setup.schema_modules.output_schemas_array import (
    ArrayOutputSamplingConfig,
    ArrayOutputTypeConfig,
)
from eir.setup.schema_modules.output_schemas_image import (
    ImageOutputSamplingConfig,
    ImageOutputTypeConfig,
)
from eir.setup.schema_modules.output_schemas_sequence import (
    SequenceOutputSamplingConfig,
    SequenceOutputTypeConfig,
)
from eir.setup.schema_modules.output_schemas_survival import SurvivalOutputTypeConfig
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.setup.schema_modules.tensor_broker_schemas import TensorBrokerConfig
from eir.setup.setup_utils import get_all_optimizer_names

if TYPE_CHECKING:
    from eir.train_utils.metrics import al_cat_metric_choices, al_con_metric_choices

al_input_configs = Sequence["InputConfig"]
al_output_configs = Sequence["OutputConfig"]

al_input_type_info = Union[
    "OmicsInputDataConfig",
    "TabularInputDataConfig",
    "SequenceInputDataConfig",
    "ByteInputDataConfig",
    "ImageInputDataConfig",
    "ArrayInputDataConfig",
]

al_optimizers = tuple(Literal[i] for i in get_all_optimizer_names())

al_feature_extractor_configs = Union[
    OmicsModelConfig,
    TabularModelConfig,
    ImageModelConfig,
    SequenceModelConfig,
    ArrayModelConfig,
]

al_feature_extractor_configs_classes = Union[
    Type[OmicsModelConfig],
    Type[TabularModelConfig],
    Type[ImageModelConfig],
    Type[SequenceModelConfig],
    Type[ArrayModelConfig],
]

al_models_classes = Union[
    Type[CNNModel],
    Type[LinearModel],
    Type[LCLModel],
    Type[SimpleLCLModel],
    Type[SimpleTabularModel],
    Type[IdentityModel],
]


al_output_module_configs_classes = (
    Type[TabularOutputModuleConfig]
    | Type[SequenceOutputModuleConfig]
    | Type[ArrayOutputModuleConfig]
)

al_output_type_configs = (
    SequenceOutputTypeConfig
    | TabularOutputTypeConfig
    | ArrayOutputTypeConfig
    | ImageOutputTypeConfig
    | SurvivalOutputTypeConfig
)

al_output_module_configs = (
    TabularOutputModuleConfig | SequenceOutputModuleConfig | ArrayOutputModuleConfig
)


al_tokenizer_choices = (
    Union[
        Literal["basic_english"],
        Literal["spacy"],
        Literal["moses"],
        Literal["toktok"],
        Literal["revtok"],
        Literal["subword"],
        Literal["bpe"],
        None,
    ],
)

al_max_sequence_length = Union[int, Literal["max", "average"]]


@dataclass
class BasicExperimentConfig:
    """
    :param output_folder:
        What to name the experiment and output folder where results are saved.

    :param n_epochs:
        Number of epochs for training.

    :param batch_size:
        Size of batches during training.

    :param valid_size:
        Size if the validation set, if float then uses a percentage. If int, then raw
        counts.

    :param manual_valid_ids_file:
        File with IDs of those samples to manually use as the validation set.
        Should be one ID per line in the file.

    :param dataloader_workers:
        Number of workers for multiprocess training and validation data loading.

    :param device:
        Device to run the training on (e.g. 'cuda:0' / 'cpu' / 'mps'). 'mps' is
        currently experimental, and might not work for all models.

    :param memory_dataset:
        Whether to load all sample into memory during training.
    """

    output_folder: str
    n_epochs: int = 10
    batch_size: int = 64
    valid_size: Union[float, int] = 0.1
    manual_valid_ids_file: Union[str, None] = None
    dataloader_workers: int = 0
    device: str = "cpu"
    memory_dataset: bool = False


@dataclass
class GlobalModelConfig:
    """
    :param compile_model:
        Whether to compile the model before training. This can be useful to speed
        up training, but may not work for all models.

    :param n_iter_before_swa:
        Number of iterations to run before activating Stochastic Weight Averaging (SWA).

    :param amp:
        Whether to use Automatic Mixed Precision. Currently only supported when
        training on GPUs.

    :param pretrained_checkpoint:
        Path to a pretrained checkpoint model file (under saved_models/ in the
        experiment output folder) to load and use as a starting point for training.

    :param strict_pretrained_loading:
        Whether to enforce that the loaded pretrained model exactly the same
        architecture as the current model. If False, will only load the layers that
        match between the two models.
    """

    compile_model: bool = False
    n_iter_before_swa: Union[None, int] = None
    amp: bool = False
    pretrained_checkpoint: Union[None, str] = None
    strict_pretrained_loading: bool = True


@dataclass
class OptimizationConfig:
    """
    :param optimizer:
        What optimizer to use.

    :param lr:
        Base learning rate for optimizer.

    :param lr_lb:
        Lower bound for learning rate when using LR scheduling

    :param b1:
        Decay of first order momentum of gradient for relevant optimizers.

    :param b2:
        Decay of second order momentum of gradient for relevant optimizers.

    :param wd:
        Weight decay.

    :param gradient_clipping:
        Max norm used for gradient clipping, with p=2.

    :param gradient_accumulation_steps:
        Number of steps to use for gradient accumulation.

    :param gradient_noise:
        Gradient noise to inject during training.
    """

    optimizer: str = "adamw"
    lr: float = 3e-04
    lr_lb: float = 0.0
    b1: float = 0.9
    b2: float = 0.999
    wd: float = 1e-04
    gradient_clipping: float = 1.0
    gradient_accumulation_steps: Union[None, int] = None
    gradient_noise: float = 0.0


@dataclass
class LRScheduleConfig:
    """
    :param find_lr:
        Whether to perform a range test of different learning rates, with the lower
        limit being what is passed in for the --lr flag. Produces a plot and exits
        with status 0 before training if this flag is active.

    :param lr_schedule:
        Whether to use cyclical, cosine or reduce on plateau learning rate schedule.
        Otherwise, keeps same learning rate

    :param lr_plateau_patience:
        Number of validation performance steps without improvement over best
        performance before reducing LR (only relevant when --lr_schedule is 'plateau'.

    :param lr_plateau_factor:
        Factor to reduce LR when running with plateau schedule.

    :param warmup_steps:
        How many steps to use in warmup. If not set, will automatically compute the
        number of steps if using an adaptive optimizer, otherwise use 2000.

    :param plot_lr_schedule:
        Whether to run LR search, plot the results and exit with status 0.
    """

    find_lr: bool = False
    lr_schedule: Literal["cycle", "plateau", "same", "cosine"] = "plateau"
    lr_plateau_patience: int = 10
    lr_plateau_factor: float = 0.2
    warmup_steps: Union[Literal["auto"], int] = "auto"
    plot_lr_schedule: bool = False


@dataclass
class TrainingControlConfig:
    """
    :param early_stopping_patience:
        Number of validation performance steps without improvement over best
        performance before terminating run.

    :param early_stopping_buffer:
        Number of iterations to run before activating early stopping checks, useful
        if networks take a while to 'kick into gear'.

    :param weighted_sampling_columns:
        Target column to apply weighted sampling on, relevant for supervised
        (tabular) targets.
        Only applies to categorical columns.
        Passing in 'all' here will use an average of all the target columns.

    :param mixing_alpha:
        Alpha parameter used for mixing (higher means more mixing).
    """

    early_stopping_patience: int = 10
    early_stopping_buffer: Union[None, int] = None
    weighted_sampling_columns: Union[None, Sequence[str]] = None
    mixing_alpha: float = 0.0


@dataclass
class EvaluationCheckpointConfig:
    """
    :param sample_interval:
        Iteration interval to perform validation and possibly attribution analysis
         if set.

    :param saved_result_detail_level:
        Level of detail to save in the results file. Higher levels will save more
        information, but will take up more space and might be slow especially
        when using many tabular targets. The details are as follows
        (each step skips a level of detail in addition to the previous):
        - 5: The default, save all metrics and plots.
        - 4: Skip validation plots and generated samples under the 'samples' folder.
        - 3: Skip plots for individual targets.
        - 2: Skip individual target plots (e.g. R2 training curve for regression
        targets).
        - 1: Skip individual target metrics (including loss).

    :param checkpoint_interval:
        Iteration interval to checkpoint (i.e. save) model.

    :param n_saved_models:
        Number of top N models to saved during training.
    """

    sample_interval: int = 200
    saved_result_detail_level: int = 5
    checkpoint_interval: Union[None, int] = None
    n_saved_models: int = 1


@dataclass
class AttributionAnalysisConfig:
    """
    :param compute_attributions:
        Whether to compute attributions / feature importance scores
        (using integrated gradients) assigned by the model with respect to the
        input features.

    :param max_attributions_per_class:
        Maximum number of samples per class to gather for
        attribution / feature importance
        analysis. Good to use when modelling on imbalanced data.

    :param attributions_every_sample_factor:
        Controls whether the attributions / feature importance
        values are computed at every
        sample interval (=1), every other sample interval (=2), etc.
        Useful when computing the attributions takes a long time, and
        we don't want to do it every time we evaluate.

    :param attribution_background_samples:
        Number of samples to use for the background in attribution / feature importance
         computations.
    """

    compute_attributions: bool = False
    max_attributions_per_class: Union[None, int] = None
    attributions_every_sample_factor: int = 1
    attribution_background_samples: int = 256


@dataclass
class SupervisedMetricsConfig:
    """

    :param cat_metrics:
        Which metrics to calculate for categorical targets.
        If not set, will use the default metrics for the task type.

    :param con_metrics:
        Which metrics to calculate for continuous targets.
        If not set, will use the default metrics for the task type.

    :param cat_averaging_metrics:
        Which metrics to use for averaging categorical targets. If not set, will use
        the default metrics for the task type.

    :param con_averaging_metrics:
        Which metrics to use for averaging continuous targets. If not set, will use
        the default metrics for the task type.
    """

    cat_metrics: "al_cat_metric_choices" = ("mcc", "acc", "roc-auc-macro", "ap-macro")
    con_metrics: "al_con_metric_choices" = ("r2", "pcc", "loss")

    cat_averaging_metrics: Optional["al_cat_metric_choices"] = None
    con_averaging_metrics: Optional["al_con_metric_choices"] = None


@dataclass
class VisualizationLoggingConfig:
    """
    :param plot_skip_steps:
        How many iterations to skip in plots.
    :param no_pbar:
        Whether to not use progress bars. Useful when stdout/stderr is written to files.
    :param log_level:
        Logging level to use. Can be one of 'debug', 'info', 'warning', 'error',
         'critical'.
    """

    plot_skip_steps: int = 200
    no_pbar: bool = False
    log_level: Literal["debug", "info", "warning", "error", "critical"] = "info"


@dataclass
class GlobalConfig:
    """
    Global configurations that are common / relevant for the whole experiment to run.
    """

    basic_experiment: BasicExperimentConfig
    model: GlobalModelConfig
    optimization: OptimizationConfig
    lr_schedule: LRScheduleConfig
    training_control: TrainingControlConfig
    evaluation_checkpoint: EvaluationCheckpointConfig
    attribution_analysis: AttributionAnalysisConfig
    metrics: SupervisedMetricsConfig
    visualization_logging: VisualizationLoggingConfig
    latent_sampling: Optional[LatentSamplingConfig] = None

    be: BasicExperimentConfig = field(init=False, repr=False)
    m: GlobalModelConfig = field(init=False, repr=False)
    opt: OptimizationConfig = field(init=False, repr=False)
    lr: LRScheduleConfig = field(init=False, repr=False)
    tc: TrainingControlConfig = field(init=False, repr=False)
    ec: EvaluationCheckpointConfig = field(init=False, repr=False)
    aa: AttributionAnalysisConfig = field(init=False, repr=False)
    met: SupervisedMetricsConfig = field(init=False, repr=False)
    vl: VisualizationLoggingConfig = field(init=False, repr=False)
    ls: Optional[LatentSamplingConfig] = field(init=False, repr=False)

    def __post_init__(self):
        self.be = self.basic_experiment
        self.m = self.model
        self.opt = self.optimization
        self.lr = self.lr_schedule
        self.tc = self.training_control
        self.ec = self.evaluation_checkpoint
        self.aa = self.attribution_analysis
        self.met = self.metrics
        self.vl = self.visualization_logging
        self.ls = self.latent_sampling


@dataclass
class InputConfig:
    """
    :param input_info:
        Information about the input source, name and type.

    :param input_type_info:
       Information specific to the input type, e.g. some augmentations are only relevant
       for omics input. Another example is the type of model to apply to the input.

    :param model_config:
        Configuration for the chosen model (i.e. feature extractor) for this input.

    :param pretrained_config:
        Configuration for using leveraging pretraining from a previous experiment.

    :param interpretation_config:
        Configuration for interpretation analysis when applicable.

    :param tensor_broker_config:
        Configuration for tensor broker when applicable. Note that this is an
        experimental feature.

    """

    input_info: "InputDataConfig"
    input_type_info: Union[
        "OmicsInputDataConfig",
        "TabularInputDataConfig",
        "SequenceInputDataConfig",
        "ByteInputDataConfig",
        "ImageInputDataConfig",
        "ArrayInputDataConfig",
    ]
    model_config: al_feature_extractor_configs
    pretrained_config: Union[None, "BasicPretrainedConfig"] = None
    interpretation_config: Union[None, "BasicInterpretationConfig"] = None
    tensor_broker_config: Union[None, TensorBrokerConfig] = None


@dataclass
class InputDataConfig:
    """
    :param input_source:
        Where on the filesystem to locate the input.

    :param input_name:
        Name to identify the input.

    :param input_type:
        Type of the input.

    :param input_inner_key:
        Inner key to use for the input. Only used when input_source is a deeplake
        dataset.
    """

    input_source: str
    input_name: str
    input_type: Literal["omics", "tabular", "sequence", "image", "bytes", "array"]
    input_inner_key: Union[None, str] = None


@dataclass
class OmicsInputDataConfig:
    """
    :param snp_file:
        Path to the relevant ``.bim`` file, used for attribution analysis.

    :param subset_snps_file:
        Path to a file with corresponding SNP IDs to subset from the main
        arrays for the modelling. Requires the ``snp_file`` parameter to
        be passed in.

    :param na_augment_alpha:
        Used to control the extent of missing data augmentation in the omics data.
        A value is sampled from a beta distribution, and the sampled value is used
        to set a percentage of the SNPs to be 'missing'.

        The alpha (α) parameter of the beta distribution, influencing the shape of the
        distribution towards 1. Higher values of alpha (compared to beta) bias the
        distribution to sample larger percentages of SNPs to be set as 'missing',
        leading to a higher likelihood of missingness.
        Conversely, lower values of alpha (compared to beta) result in sampling lower
        percentages, thus reducing the probability and extent of missingness.
        For example, setting alpha to 1.0 and beta to 5.0 will skew the distribution
        towards lower percentages of missingness, since beta is significantly larger.
        Setting alpha to 5.0 and beta to 1.0 will skew the distribution towards higher
        percentages of missingness, since alpha is significantly larger.
        Examples:
        - alpha = 1.0, beta = 9.0:  μ=E(X)=0.05, σ=SD(X)=0.0476 (avg 5% missing)
        - alpha = 1.0, beta = 4.0:  μ=E(X)=0.2, σ=SD(X)=0.1633 (avg 20% missing)

    :param na_augment_beta:
        Used to control the extent of missing data augmentation in the omics data.
        A value is sampled from a beta distribution, and the sampled value is used
        to set a percentage of the SNPs to be 'missing'.

        Beta (β) parameter of the beta distribution, influencing the shape of the
        distribution towards 0. Higher values of beta (compared to alpha) bias the
        distribution to sample smaller percentages of SNPs to be set as 'missing',
        leading to a lower likelihood and extent of missingness.
        Conversely, lower values of beta (compared to alpha) result in sampling
        larger percentages, thus increasing the probability and extent of missingness.

    :param shuffle_augment_alpha:
        Used to control the extent of shuffling data augmentation in the omics data.
        A value is sampled from a beta distribution, and the sampled value is used to
        determine the percentage of the SNPs to be shuffled.

        The alpha (α) parameter of the beta distribution, influencing the shape of
        the distribution towards 1. Higher values of alpha (compared to beta) bias
        the distribution to sample larger percentages of SNPs to be shuffled, leading
        to a higher likelihood of extensive shuffling. Conversely, lower values of
        alpha (compared to beta) result in sampling lower percentages, thus reducing
        the extent of shuffling. Setting alpha to a significantly larger value than
        beta will skew the distribution towards higher percentages of shuffling.
        Examples:
        - alpha = 1.0, beta = 9.0:  μ=E(X)=0.05, σ=SD(X)=0.0476 (avg 5% shuffled)
        - alpha = 1.0, beta = 4.0:  μ=E(X)=0.2, σ=SD(X)=0.1633 (avg 20% shuffled)

    :param shuffle_augment_beta:
        Used to control the extent of shuffling data augmentation in the omics data.
        A value is sampled from a beta distribution, and the sampled value is used to
        determine the percentage of the SNPs to be shuffled.

        Beta (β) parameter of the beta distribution, influencing the shape of the
        distribution towards 0. Higher values of beta (compared to alpha) bias the
        distribution to sample smaller percentages of SNPs to be shuffled, leading to
        a lower likelihood and extent of shuffling. Conversely, lower values of beta
        (compared to alpha) result in sampling larger percentages, thus increasing
        the likelihood and extent of shuffling.

    :param omics_format:
        Currently unsupported (i.e. does nothing), which format the omics data is in.

    :param mixing_subtype:
        Which type of mixing to use on the omics data given that ``mixing_alpha`` is
        set >0.0 in the global configuration.

    :param modality_dropout_rate:
        Dropout rate to apply to the modality, e.g. 0.2 means that 20% of the time,
        this modality will be dropped out during training.
    """

    snp_file: Optional[str] = None
    subset_snps_file: Optional[str] = None
    na_augment_alpha: float = 1.0
    na_augment_beta: float = 5.0
    shuffle_augment_alpha: float = 0.0
    shuffle_augment_beta: float = 0.0
    omics_format: Literal["one-hot"] = "one-hot"
    mixing_subtype: Union[Literal["mixup", "cutmix-block", "cutmix-uniform"]] = "mixup"
    modality_dropout_rate: float = 0.0


@dataclass
class TabularInputDataConfig:
    """
    :param input_cat_columns:
        Which columns to use as a categorical inputs from the ``input_source`` specified
        in the ``input_info`` field of the relevant ``.yaml``.

    :param input_con_columns:
        Which columns to use as a continuous inputs from the ``input_source`` specified
        in the ``input_info`` field of the relevant ``.yaml``.

    :param label_parsing_chunk_size:
        Number of rows to process at time when loading in the ``input_source``. Useful
        when RAM is limited.

    :param mixing_subtype:
        Which type of mixing to use on the tabular data given that ``mixing_alpha`` is
        set >0.0 in the global configuration.

    :param modality_dropout_rate:
        Dropout rate to apply to the modality, e.g. 0.2 means that 20% of the time,
        this modality will be dropped out during training.
    """

    input_cat_columns: Sequence[str] = field(default_factory=list)
    input_con_columns: Sequence[str] = field(default_factory=list)
    label_parsing_chunk_size: Union[None, int] = None
    mixing_subtype: Literal["mixup"] = "mixup"
    modality_dropout_rate: float = 0.0


@dataclass
class SequenceInputDataConfig:
    """
    :param vocab_file:
        An optional text file containing pre-defined vocabulary to use
        for the training. If this is not passed in, the framework will automatically
        build the vocabulary from the training data. Passing in a vocabulary file is
        therefore useful if (a) you want to manually specify / limit the vocabulary used
        and/or (b) you want to save time by pre-computing the vocabulary.

    :param max_length:
        Maximum length to truncate/pad sequences to. This can be an integer or the
        values 'max' or 'average'. The 'max' keyword will use the maximum sequence
        length found in the training data, while the 'average' will use the average
        length across all training samples.

    :param sampling_strategy_if_longer:
        Controls how sequences are truncated if they are longer than the specified
        ``max_length`` parameter. Using 'from_start' will always truncate from the
        beginning of the sequence, ensuring the the samples will always be the same
        during training. Setting this parameter to ``uniform`` will uniformly sample
        a slice of a given sample sequence during training. Note that for consistency,
        the validation/test set samples always use the ``from_start`` setting when
        truncating.

    :param min_freq:
        Minimum number of times a token must appear in the total training data to be
        included in the vocabulary. Note that this setting will not do anything if
        passing in ``vocab_file``.

    :param split_on:
        Which token to split the sequence on to generate separate tokens for the
        vocabulary.

    :param tokenizer:
        Which tokenizer to use. Relevant if modelling on language, but not as much when
        doing it on other arbitrary sequences.

    :param tokenizer_language:
        Which language rules the tokenizer should apply when tokenizing the raw data.

    :param adaptive_tokenizer_max_vocab_size:
        If using an adaptive tokenizer ("bpe"), this parameter controls the maximum
        size of the vocabulary.

    :param mixing_subtype:
        Which type of mixing to use on the sequence data given that ``mixing_alpha`` is
        set >0.0 in the global configuration.

    :param modality_dropout_rate:
        Dropout rate to apply to the modality, e.g. 0.2 means that 20% of the time,
        this modality will be dropped out during training.
    """

    vocab_file: Union[None, str] = None
    max_length: al_max_sequence_length = "average"
    sampling_strategy_if_longer: Literal["from_start", "uniform"] = "uniform"
    min_freq: int = 10
    split_on: Optional[str] = " "
    tokenizer: al_tokenizer_choices = None  # type: ignore
    tokenizer_language: Union[str, None] = None
    adaptive_tokenizer_max_vocab_size: Optional[int] = None
    mixing_subtype: Literal["mixup"] = "mixup"
    modality_dropout_rate: float = 0.0


@dataclass
class BasicPretrainedConfig:
    """
    :param model_path:
        Path save model from an EIR training run to load. Note that currently this
        only supports if it's in a folder from an EIR training run, as the current
        prototype functionality uses that to e.g. find configuration from the
        pretrained model run.

    :param load_module_name:
        Name of the module to extract and use in the respective input feature
        extraction.
    """

    model_path: str
    load_module_name: str


@dataclass
class BasicInterpretationConfig:
    """
    :param interpretation_sampling_strategy:
        How to sample sequences for attribution analysis. `first_n` always grabs the
        same first n values from the beginning of the dataset to interpret, while
        `random_sample` will sample uniformly from the whole dataset without
        replacement.

    :param num_samples_to_interpret:
        How many samples to interpret.

    :param manual_samples_to_interpret:
        IDs of samples to always interpret, irrespective of
        `interpretation_sampling_strategy` and `num_samples_to_interpret`. A caveat
        here is that they must be present in the dataset that is being interpreted
        (e.g. validation / test dataset), meaning that adding IDs here that happen to
        be in the training dataset will not work.
    """

    interpretation_sampling_strategy: Literal["first_n", "random_sample"] = "first_n"
    num_samples_to_interpret: int = 10
    manual_samples_to_interpret: Union[Sequence[str], None] = None


@dataclass
class ByteInputDataConfig:
    """
    :param byte_encoding:
        Which byte encoding to use when reading the binary data, currently only
        support uint8.

    :param max_length:
        Maximum length to truncate/pad sequences to. While in sequence models this
        generally refers to words, here we are referring to number of bytes.

    :param sampling_strategy_if_longer:
        Controls how sequences are truncated if they are longer than the specified
        ``max_length`` parameter. Using 'from_start' will always truncate from the
        beginning of the byte sequence, ensuring the the samples will always be the same
        during training. Setting this parameter to ``uniform`` will uniformly sample
        a slice of a given sample sequence during training. Note that for consistency,
        the validation/test set samples always use the ``from_start`` setting when
        truncating.

    :param mixing_subtype:
        Which type of mixing to use on the bytes data given that ``mixing_alpha`` is
        set >0.0 in the global configuration.

    :param modality_dropout_rate:
        Dropout rate to apply to the modality, e.g. 0.2 means that 20% of the time,
        this modality will be dropped out during training.
    """

    max_length: int = 256
    byte_encoding: Literal["uint8"] = "uint8"
    sampling_strategy_if_longer: Literal["from_start", "uniform"] = "uniform"
    mixing_subtype: Literal["mixup"] = "mixup"
    modality_dropout_rate: float = 0.0


@dataclass
class ImageInputDataConfig:
    """
    :param auto_augment:
        Setting this to True will use TrivialAugment Wide augmentation.

    :param size:
        Target size of the images for training.  If size is a sequence like
        (h, w), output size will be matched to this. If size is an int,
        the image will be resized to (size, size).

    :param resize_approach:
        The method used for resizing the images. Options are:
        - "resize": Directly resize the image to the target size.
        - "randomcrop": Resize the image to a larger size than the target and then
        apply a random crop to the target size.
        - "centercrop": Resize the image to a larger size than the target and then
        apply a center crop to the target size.

    :param adaptive_normalization_max_samples:
        If using adaptive normalization (channel),
        how many samples to use to compute the normalization parameters.
        If None, will use all samples.

    :param mean_normalization_values:
        Average channel values to normalize images with. This can be a sequence matching
        the number of channels, or None. If None and using a pretrained model, the
        values used for the model pretraining will be used. If None and training from
        scratch, will iterate over training data and compute the running average
        per channel.

    :param stds_normalization_values:
        Standard deviation channel values to normalize images with. This can be a
        sequence mathing the number of channels, or None. If None and using a
        pretrained model, the values used for the model pretraining will be used.
        If None and training from scratch, will iterate over training data and compute
        the running average per channel.

    :param mode:
        An explicit mode to convert loaded images to. Useful when working with
        input data with a mixed number of channels, or you want to convert
        images to a specific mode.
        Options are
        - "RGB": Red, Green, Blue (channels=3)
        - "L": Grayscale (channels=1)
        - "RGBA": Red, Green, Blue, Alpha (channels=4)

    :param num_channels:
        Number of channels in the images. If None, will try to infer the number of
        channels from a random image in the training data. Useful when known
        ahead of time how many channels the images have, will raise an error if
        an image with a different number of channels is encountered.

    :param mixing_subtype:
        Which type of mixing to use on the image data given that ``mixing_alpha`` is
        set >0.0 in the global configuration.

    :param modality_dropout_rate:
        Dropout rate to apply to the modality, e.g. 0.2 means that 20% of the time,
        this modality will be dropped out during training.
    """

    auto_augment: bool = True
    size: Sequence[int] = (64,)
    resize_approach: Union[Literal["resize", "randomcrop", "centercrop"]] = "resize"
    adaptive_normalization_max_samples: Optional[int] = None
    mean_normalization_values: Union[None, Sequence[float]] = None
    stds_normalization_values: Union[None, Sequence[float]] = None
    mode: Optional[Literal["RGB", "L", "RGBA"]] = None
    num_channels: Optional[int] = None
    mixing_subtype: Union[Literal["mixup"], Literal["cutmix"]] = "mixup"
    modality_dropout_rate: float = 0.0


@dataclass
class ArrayInputDataConfig:
    """
    :param mixing_subtype:
        Which type of mixing to use on the image data given that ``mixing_alpha`` is
        set >0.0 in the global configuration.

    :param modality_dropout_rate:
        Dropout rate to apply to the modality, e.g. 0.2 means that 20% of the time,
        this modality will be dropped out during training.

    :param normalization:
        Which type of normalization to apply to the array data. If ``element``, will
        normalize each element in the array independently. If ``channel``, will
        normalize each channel in the array independently.
        For 'channel', assumes PyTorch format where the channel dimension is the
        first dimension.

    :param adaptive_normalization_max_samples:
        If using adaptive normalization (channel / element),
        how many samples to use to compute the normalization parameters.
        If None, will use all samples.
    """

    mixing_subtype: Union[Literal["mixup"]] = "mixup"
    modality_dropout_rate: float = 0.0
    normalization: Optional[Literal["element", "channel"]] = "channel"
    adaptive_normalization_max_samples: Optional[int] = None


@dataclass
class FusionConfig:
    """
    :param model_type:
        Which type of fusion model to use.

    :param model_config:
        Fusion model configuration.
    """

    model_type: Literal["mlp-residual", "identity", "mgmoe", "pass-through"]
    model_config: Union[ResidualMLPConfig, IdentityConfig, MGMoEModelConfig]
    tensor_broker_config: Union[None, TensorBrokerConfig] = None


@dataclass
class OutputInfoConfig:
    """
    :param output_source:
        Where on the filesystem to locate the output (if applicable)

    :param output_name:
        Name to identify the output.

    :param output_type:
        Type of the output.
    """

    output_source: str
    output_name: str
    output_type: Literal["tabular", "sequence", "array"]
    output_inner_key: Optional[str] = None


@dataclass
class OutputConfig:
    """
    :param output_info:
        Information about the output source, name and type.

    :param output_type_info:
       Information specific to the output type, e.g. which columns to predict
       from a tabular file.

    :param model_config:
        Configuration for the chosen model (i.e. output module after fusion) for this
        output.

    :param sampling_config:
        Configuration for how to sample results from the output module.
    """

    output_info: OutputInfoConfig
    output_type_info: (
        TabularOutputTypeConfig
        | SequenceOutputTypeConfig
        | ArrayOutputTypeConfig
        | ImageOutputTypeConfig
        | SurvivalOutputTypeConfig
    )
    model_config: (
        TabularOutputModuleConfig | SequenceOutputModuleConfig | ArrayOutputModuleConfig
    )

    sampling_config: Optional[
        SequenceOutputSamplingConfig
        | ArrayOutputSamplingConfig
        | ImageOutputSamplingConfig
        | dict
    ] = None
    tensor_broker_config: Union[None, TensorBrokerConfig] = None



from pathlib import Path

from eir.experiment_io.label_transformer_io import load_transformers
from eir.experiment_io.output_object_io_modules.output_io_utils import (
    load_output_config_from_yaml,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
    set_up_tabular_output,
)


def load_tabular_output_object(
    serialized_output_folder: Path,
    run_folder: Path,
) -> ComputedTabularOutputInfo:
    config_path = serialized_output_folder / "output_config.yaml"

    transformers = load_transformers(run_folder=run_folder)

    output_config = load_output_config_from_yaml(output_config_path=config_path)

    loaded_object = set_up_tabular_output(
        output_config=output_config,
        target_transformers=transformers,
    )

    return loaded_object



from pathlib import Path

import numpy as np

from eir.experiment_io.label_transformer_io import load_transformers
from eir.experiment_io.output_object_io_modules.output_io_utils import (
    load_output_config_from_yaml,
)
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
    set_up_survival_output,
)


def load_survival_output_object(
    serialized_output_folder: Path,
    run_folder: Path,
) -> ComputedSurvivalOutputInfo:
    config_path = serialized_output_folder / "output_config.yaml"

    transformers = load_transformers(run_folder=run_folder)

    output_config = load_output_config_from_yaml(output_config_path=config_path)

    baseline_hazard = None
    if (serialized_output_folder / "baseline_hazard.npy").exists():
        baseline_hazard = np.load(serialized_output_folder / "baseline_hazard.npy")

    baseline_unique_times = None
    if (serialized_output_folder / "baseline_unique_times.npy").exists():
        baseline_unique_times = np.load(
            serialized_output_folder / "baseline_unique_times.npy"
        )

    loaded_object = set_up_survival_output(
        output_config=output_config,
        target_transformers=transformers,
        baseline_hazard=baseline_hazard,
        baseline_unique_times=baseline_unique_times,
    )

    return loaded_object



from typing import TYPE_CHECKING

from eir.train_utils.evaluation_modules.train_handlers_sequence_output import (
    sequence_out_single_sample_evaluation_wrapper,
)

if TYPE_CHECKING:
    from eir.predict import PredictExperiment


def predict_sequence_wrapper(
    predict_experiment: "PredictExperiment",
    output_folder: str,
) -> None:
    sequence_out_single_sample_evaluation_wrapper(
        experiment=predict_experiment,
        iteration=0,
        input_objects=predict_experiment.inputs,
        auto_dataset_to_load_from=predict_experiment.test_dataset,
        output_folder=output_folder,
    )



from typing import Iterable, Optional, Sequence, Type, Union

import torch
from torch_optimizer import _NAME_OPTIM_MAP
from tqdm import tqdm
from transformers.models.auto.modeling_auto import MODEL_MAPPING_NAMES

from eir.utils.logging import get_logger

al_collector_classes = Union[
    Type["ChannelBasedRunningStatistics"], Type["ElementBasedRunningStatistics"]
]

logger = get_logger(name=__name__)


class ChannelBasedRunningStatistics:
    """
    Adapted from https://gist.github.com/thomasbrandon/ad5b1218fc573c10ea4e1f0c63658469.

    Note assumes CxHxW input shape. final_n_dims is the number of last dimensions
    that are averaged over. For example, if the input shape is (C, H, W) and
    final_n_dims is 2, then the mean and variance will be computed over the
    (H, W) dimensions, resulting in tensors of shape (C,).
    """

    def __init__(self, final_n_dims: int = 2):
        self.final_n_dims: int = final_n_dims
        self.shape: tuple = tuple()
        self.n: int = 0
        self.sum: torch.Tensor = torch.empty(0)
        self.num_var: torch.Tensor = torch.empty(0)

    @torch.no_grad()
    def update(self, data: torch.Tensor):
        data = data.reshape(*list(data.shape[: -self.final_n_dims]) + [-1])

        mask = ~torch.isnan(data)
        masked_data = data.clone()
        masked_data[~mask] = 0.0

        new_n: int = data.shape[-1]
        new_var: torch.Tensor = masked_data.var(-1)
        new_sum: torch.Tensor = masked_data.sum(-1)

        if self.n == 0:
            self.n = new_n
            self.shape = data.shape[:-1]
            self.sum = new_sum
            self.num_var = new_var.mul_(new_n)
        else:
            ratio = self.n / new_n
            t = (self.sum / ratio).sub_(new_sum).pow_(2)

            self.num_var.add_(other=new_var, alpha=new_n)
            self.num_var.add_(other=t, alpha=ratio / (self.n + new_n))

            self.sum += new_sum
            self.n += new_n

    @property
    def mean(self) -> torch.Tensor:
        return self.sum / self.n

    @property
    def var(self) -> torch.Tensor:
        return self.num_var / self.n

    @property
    def std(self) -> torch.Tensor:
        return self.var.sqrt()


class ElementBasedRunningStatistics:
    def __init__(self, shape: tuple):
        self.shape = shape
        self.n = 0
        self._mean = torch.zeros(shape)
        self._ssdm = torch.zeros(shape)

    @torch.no_grad()
    def update(self, x: torch.Tensor):
        mask = ~torch.isnan(x)
        if not mask.any():
            return

        x_masked = x.clone()
        x_masked[~mask] = self._mean[~mask]

        self.n += 1
        delta = x_masked - self._mean
        self._mean += delta / self.n
        delta2 = x_masked - self._mean
        self._ssdm += delta * delta2

    @property
    def mean(self) -> torch.Tensor:
        return self._mean

    @property
    def var(self) -> torch.Tensor:
        return self._ssdm / (self.n - 1) if self.n > 1 else torch.square(self._mean)

    @property
    def std(self) -> torch.Tensor:
        return torch.sqrt(self.var)


def collect_stats(
    tensor_iterable: Iterable[torch.Tensor],
    collector_class: al_collector_classes,
    shape: tuple,
    max_samples: Optional[int] = None,
    name: Optional[str] = None,
) -> ChannelBasedRunningStatistics | ElementBasedRunningStatistics:
    stats = set_up_collector_instance(collector_class=collector_class, shape=shape)

    for index, it in enumerate(tqdm(tensor_iterable, desc="Gathering Statistics: ")):
        if hasattr(it, "data"):
            stats.update(it.data)
        else:
            stats.update(it)

        if max_samples is not None and index >= max_samples:
            break

    if name is not None:
        logger.info(
            f"Collected stats for {name}, "
            f"with means {stats.mean} and standard deviations {stats.std}."
        )

    return stats


def set_up_collector_instance(
    collector_class: al_collector_classes,
    shape: tuple,
) -> ChannelBasedRunningStatistics | ElementBasedRunningStatistics:
    stats: ChannelBasedRunningStatistics | ElementBasedRunningStatistics
    if collector_class is ChannelBasedRunningStatistics:
        final_n_dims = len(shape) - 1
        stats = ChannelBasedRunningStatistics(final_n_dims=final_n_dims)
    elif collector_class is ElementBasedRunningStatistics:
        stats = ElementBasedRunningStatistics(shape=shape)
    else:
        raise ValueError(
            f"collector_class must be either ChannelBasedRunningStatistics or "
            f"ElementBasedRunningStatistics, "
            f"got {collector_class}."
        )
    return stats


def get_base_optimizer_names() -> set:
    base_names = {
        "sgdm",
        "adam",
        "adamw",
        "adahessian",
        "adabelief",
        "adabeliefw",
        "soap",
    }

    return base_names


def get_all_optimizer_names() -> Sequence[str]:
    external_optimizers = set(_NAME_OPTIM_MAP.keys())
    base_optimizers = get_base_optimizer_names()
    all_optimizers = set.union(base_optimizers, external_optimizers)
    all_optimizers_list = sorted(list(all_optimizers))

    return all_optimizers_list


def get_all_hf_model_names() -> Sequence[str]:
    all_models = sorted(list(MODEL_MAPPING_NAMES.keys()))
    unsupported = get_unsupported_hf_models()
    unsupported_names = unsupported.keys()
    return [i for i in all_models if i not in unsupported_names]


def get_unsupported_hf_models() -> dict:
    unsupported = {
        "audio-spectrogram-transformer": "Not strictly sequence model.",
        "align": "Not strictly sequence model.",
        "altclip": "Not strictly sequence model.",
        "aria": "Too large for testing.",
        "aria_text": "Too large for testing.",
        "autoformer": "Not strictly sequence model.",
        "bark": "Not strictly sequence model.",
        "bamba": "Too large for testing.",
        "beit": "Not strictly sequence model.",
        "bit": "Not strictly sequence model.",
        "blip": "Not strictly sequence model.",
        "blip-2": "Not strictly sequence model.",
        "bridgetower": "Not strictly sequence model.",
        "bros": "Not strictly sequence model.",
        "canine": "Cannot do straightforward look up of embeddings.",
        "clap": "Not strictly sequence model.",
        "clip": "Not strictly sequence model.",
        "clipseg": "Not strictly sequence model.",
        "clip_vision_model": "Not strictly sequence model.",
        "clip_text_model": "CLIPTextModel.forward() got an unexpected keyword argument"
        " 'inputs_embeds'",
        "clvp": "Not strictly sequence model.",
        "chameleon": "Not strictly sequence model.",
        "chinese_clip": "Not strictly sequence model.",
        "chinese_clip_vision_model": "Not strictly sequence model.",
        "conditional_detr": "Not strictly sequence model.",
        "convbert": "HF error.",
        "convnext": "Not strictly sequence model..",
        "convnextv2": "Not strictly sequence model..",
        "cpmant": "AttributeError: 'NoneType' object has no attribute 'dtype'",
        "cvt": "Not strictly sequence model.",
        "dac": "Not strictly sequence model.",
        "data2vec-audio": "Not strictly sequence model.",
        "data2vec-vision": "Not strictly sequence model.",
        "decision_transformer": "Raises NotImplementedError",
        "decision_transformer_gpt2": "Raises ValueError: Unrecognized model identifier",
        "deformable_detr": "Not strictly sequence model.",
        "deit": "Not strictly sequence model.",
        "deta": "Not strictly sequence model.",
        "dinat": "Not strictly sequence model.",
        "dinov2": "Not strictly sequence model.",
        "dinov2_with_registers": "Not strictly sequence model.",
        "detr": "Not strictly sequence model.",
        "donut-swin": "Not strictly sequence model.",
        "dpr": "Not strictly sequence model.",
        "dpt": "Not strictly sequence model.",
        "encodec": "Not strictly sequence model.",
        "esm": "Needs fixing/special handling of init parameters.",
        "efficientformer": "Not strictly sequence model.",
        "efficientnet": "Not strictly sequence model.",
        "ernie_m": "AttributeError: 'bool' object has no attribute 'to'.",
        "fastspeech2_conformer": "Not strictly sequence model.",
        "fsmt": "Not strictly sequence model.",
        "focalnet": "Not strictly sequence model.",
        "funnel": "HF error.",
        "flava": "Not strictly sequence model.",
        "nat": "Requires the natten library",
        "glm": "The size of tensor a (2) must match the size of tensor b (64) "
        "at non-singleton dimension 3",
        "glpn": "Not strictly sequence model.",
        "granite": "Very large model, not suitable for testing.",
        "granitemoe": "Very large model, not suitable for testing.",
        "graphormer": "Not strictly sequence model.",
        "groupvit": "Not strictly sequence model.",
        "grounding-dino": "Not strictly sequence model.",
        "gpt_neo": "Configuration troublesome w.r.t. attn layers matching num_layers.",
        "gptj": "Very large model, not suitable for testing.",
        "gpt2": "Very large model, not suitable for testing.",
        "gpt-sw3": "Very large model, not suitable for testing.",
        "gpt_bigcode": "Very large model, not suitable for testing.",
        "gptsan-japanese": "AttributeError: 'MoECausalLMOutputWithPast' object "
        "has no attribute 'last_hidden_state'.",
        "hubert": "Cannot do straightforward look up of embeddings.",
        "hiera": "Not strictly sequence model.",
        "ibert": "Very large model, not suitable for testing.",
        "informer": "Not strictly sequence model.",
        "idefics": "Not strictly sequence model.",
        "idefics2": "Not strictly sequence model.",
        "idefics3": "Very large model, not suitable for testing.",
        "idefics3_vision": "Not strictly sequence model.",
        "ijepa": "Not strictly sequence model.",
        "jamba": "Fast Mamba kernels are not available error.",
        "jukebox": "Not strictly sequence model.",
        "jetmoe": "Shape mismatch error.",
        "kosmos-2": "Not strictly sequence model.",
        "layoutlmv2": "Not strictly sequence model.",
        "layoutlmv3": "Not strictly sequence model.",
        "levit": "Not strictly sequence model.",
        "lilt": "Matmul incompatible shapes with default params.",
        "lxmert": "Not strictly sequence model.",
        "mamba2": "An issue with one input dimension being 0 in the hidden states when"
        " padding in HF forward.",
        "maskformer": "Not strictly sequence model.",
        "mask2former": "Not strictly sequence model.",
        "maskformer-swin": "Not strictly sequence model.",
        "mctct": "get_input_embeddings() raises NotImplementedError.",
        "mgp-str": "Not strictly sequence model.",
        "mimi": "Not strictly sequence model.",
        "mistral": "Size mismatch error.",
        "mobilenet_v1": "Not strictly sequence model.",
        "mobilenet_v2": "Not strictly sequence model.",
        "mobilevit": "Not strictly sequence model.",
        "mobilevitv2": "Not strictly sequence model.",
        "moonshine": "Not strictly sequence model.",
        "musicgen": "Not strictly sequence model.",
        "musicgen_melody": "Not strictly sequence model.",
        "mt5": "Not implemented in EIR for feature extraction yet.",
        "oneformer": "Not strictly sequence model.",
        "omdet-turbo": "Not strictly sequence model.",
        "open-llama": "'OpenLlamaAttention' object has no attribute 'rope_theta'.",
        "owlvit": "Not strictly sequence model.",
        "owlv2": "Not strictly sequence model.",
        "nllb": "Throws ValueError: Unrecognized model identifier: nllb",
        "retribert": "Cannot do straightforward look up of embeddings.",
        "resnet": "Not strictly sequence model.",
        "regnet": "Not strictly sequence model.",
        "rt_detr": "Not strictly sequence model.",
        "sam": "Not strictly sequence model.",
        "segformer": "Not strictly sequence model.",
        "seamless_m4t": "Not strictly sequence model.",
        "seamless_m4t_v2": "Not strictly sequence model.",
        "sew": "Not strictly sequence model.",
        "sew-d": "Not strictly sequence model.",
        "seggpt": "Not strictly sequence model.",
        "siglip": "Not strictly sequence model.",
        "siglip_vision_model": "Not strictly sequence model.",
        "speech_to_text": "Not strictly sequence model.",
        "speecht5": "Not strictly sequence model.",
        "swiftformer": "Not strictly sequence model.",
        "swin": "Not strictly sequence model.",
        "swinv2": "Not strictly sequence model.",
        "swin2sr": "Not strictly sequence model.",
        "table-transformer": "Not strictly sequence model.",
        "tapas": "TapasModel requires the torch-scatter library.",
        "time_series_transformer": "Not strictly sequence model.",
        "textnet": "Not strictly sequence model.",
        "timesformer": "Not strictly sequence model.",
        "timm_backbone": "Not strictly sequence model.",
        "timm_wrapper": "Not strictly sequence model.",
        "trajectory_transformer": "Not strictly sequence model.",
        "transfo-xl": "Internal TypeError: type_as() missing 1 required positional"
        " arguments: other.",
        "tvlt": "Not strictly sequence model.",
        "tvp": "Not strictly sequence model.",
        "patchtsmixer": "Not strictly sequence model.",
        "patchtst": "Not strictly sequence model.",
        "perceiver": "Not strictly sequence model.",
        "pixtral": "Not strictly sequence model.",
        "poolformer": "Not strictly sequence model.",
        "pvt": "Not strictly sequence model.",
        "pvt_v2": "Not strictly sequence model.",
        "qdqbert": "ImportError.",
        "udop": "Not strictly sequence model.",
        "qwen2_audio_encoder": "Not strictly sequence model.",
        "qwen2_vl": "Not strictly sequence model.",
        "unispeech": "Not strictly sequence model.",
        "unispeech-sat": "Not strictly sequence model.",
        "univnet": "Not strictly sequence model.",
        "umt5": "AttributeError: property 'num_hidden_layers' of "
        "'UMT5Config' object has no setter",
        "van": "Not strictly sequence model.",
        "videomae": "Not strictly sequence model.",
        "vilt": "Not strictly sequence model.",
        "vit": "Not strictly sequence model.",
        "vitdet": "Not strictly sequence model.",
        "vits": "Not strictly sequence model.",
        "vivit": "Not strictly sequence model.",
        "vit_hybrid": "Not strictly sequence model.",
        "vit_mae": "Not strictly sequence model.",
        "vit_msn": "Not strictly sequence model.",
        "vision-text-dual-encoder": "Not strictly sequence model.",
        "xclip": "Not strictly sequence model.",
        "xmod": "ValueError: Input language unknown. Please call "
        "`XmodPreTrainedModel.set_default_language().",
        "wav2vec2": "Not strictly sequence model.",
        "wav2vec2-conformer": "Not strictly sequence model.",
        "wav2vec2-bert": "Not strictly sequence model.",
        "wavlm": "NotImplementedError.",
        "whisper": "Not strictly sequence model.",
        "yolos": "Not strictly sequence model.",
        "zamba": "Random None is not iterable error.",
    }

    return unsupported



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Sequence, Set, Type

import polars as pl
from sklearn.preprocessing import KBinsDiscretizer, StandardScaler

from eir.data_load import label_setup
from eir.data_load.data_source_modules.deeplake_ops import is_deeplake_dataset
from eir.data_load.label_setup import (
    TabularFileInfo,
    al_label_transformers,
    transform_label_df,
)
from eir.experiment_io.label_transformer_io import load_transformers
from eir.predict_modules.predict_tabular_input_setup import prep_missing_con_dict
from eir.setup.config import Configs
from eir.setup.schemas import (
    ArrayOutputTypeConfig,
    ImageOutputTypeConfig,
    OutputConfig,
    SequenceOutputTypeConfig,
    SurvivalOutputTypeConfig,
    TabularOutputTypeConfig,
)
from eir.target_setup.target_label_setup import (
    MissingTargetsInfo,
    gather_data_pointers_from_data_source,
    gather_torch_null_missing_ids,
    get_missing_targets_info,
    get_tabular_target_file_infos,
    synchronize_missing_survival_values,
    transform_durations_with_nans,
    update_labels_df,
)
from eir.target_setup.target_setup_utils import IdentityTransformer
from eir.utils.logging import get_logger

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass
class PredictTargetLabels:
    predict_labels: pl.DataFrame
    label_transformers: Dict[str, al_label_transformers]

    @property
    def all_labels(self) -> pl.DataFrame:
        return self.predict_labels


@dataclass
class MergedPredictTargetLabels:
    predict_labels: pl.DataFrame
    label_transformers: dict[str, dict[str, al_label_transformers]]
    missing_ids_per_output: MissingTargetsInfo

    @property
    def all_labels(self) -> pl.DataFrame:
        return self.predict_labels


def get_tabular_target_labels_for_predict(
    output_folder: str,
    tabular_info: TabularFileInfo,
    output_name: str,
    ids: Sequence[str],
) -> PredictTargetLabels:
    all_columns = list(tabular_info.cat_columns) + list(tabular_info.con_columns)
    if not all_columns:
        raise ValueError(f"No columns specified in {tabular_info}.")

    con_columns = list(tabular_info.con_columns)
    cat_columns = list(tabular_info.cat_columns)

    df_labels_test = _load_labels_for_predict(
        tabular_info=tabular_info,
        ids_to_keep=ids,
    )

    label_transformers = load_transformers(
        output_folder=output_folder,
        transformers_to_load={output_name: all_columns},
    )

    labels_df = parse_labels_for_predict(
        con_columns=con_columns,
        cat_columns=cat_columns,
        df_labels_test=df_labels_test,
        label_transformers=label_transformers[output_name],
    )

    labels = PredictTargetLabels(
        predict_labels=labels_df,
        label_transformers=label_transformers,
    )

    return labels


def _load_labels_for_predict(
    tabular_info: TabularFileInfo,
    ids_to_keep: Sequence[str],
) -> pl.DataFrame:
    parse_wrapper = label_setup.get_label_parsing_wrapper(
        label_parsing_chunk_size=tabular_info.parsing_chunk_size
    )
    df_labels_test = parse_wrapper(
        label_file_tabular_info=tabular_info,
        ids_to_keep=ids_to_keep,
    )

    return df_labels_test


def parse_labels_for_predict(
    con_columns: Sequence[str],
    cat_columns: Sequence[str],
    df_labels_test: pl.DataFrame,
    label_transformers: al_label_transformers,
) -> pl.DataFrame:

    all_con_transformers = _extract_target_con_transformers(
        label_transformers=label_transformers,
        con_columns=con_columns,
    )

    con_standardscaler_transformers = {
        key: value
        for key, value in all_con_transformers.items()
        if isinstance(value, StandardScaler)
    }

    train_con_column_means = prep_missing_con_dict(
        con_transformers=con_standardscaler_transformers
    )

    df_labels_test = label_setup.handle_missing_label_values_in_df(
        df=df_labels_test,
        cat_label_columns=cat_columns,
        con_label_columns=con_columns,
        con_manual_values=train_con_column_means,
        name="test_df",
        impute_missing=False,
    )

    assert len(label_transformers) > 0, "No label transformers found"

    df_labels_test = transform_label_df(
        df_labels=df_labels_test,
        label_transformers=label_transformers,
        impute_missing=False,
    )

    return df_labels_test


def _extract_target_con_transformers(
    label_transformers: al_label_transformers,
    con_columns: Sequence[str],
) -> Dict[str, StandardScaler | KBinsDiscretizer | IdentityTransformer]:
    con_transformers = {}

    for target_column, transformer_object in label_transformers.items():
        if target_column not in con_columns:
            continue

        assert target_column not in con_transformers
        assert isinstance(
            transformer_object, (StandardScaler, KBinsDiscretizer, IdentityTransformer)
        )

        con_transformers[target_column] = transformer_object

    assert len(con_transformers) == len(con_columns)

    return con_transformers


def get_target_labels_for_testing(
    configs_overloaded_for_predict: Configs,
    ids: Sequence[str],
) -> MergedPredictTargetLabels:
    pc = configs_overloaded_for_predict
    output_configs = pc.output_configs

    all_ids: set[str] = set(ids)
    per_modality_missing_ids: dict[str, set[str]] = {}
    label_transformers: dict[str, Any] = {}

    test_labels_df = pl.DataFrame(schema={"ID": pl.Utf8})

    tabular_target_files_info = get_tabular_target_file_infos(
        output_configs=pc.output_configs
    )

    for output_config in output_configs:
        output_source = output_config.output_info.output_source
        output_type_info = output_config.output_type_info
        output_name = output_config.output_info.output_name
        logger.info(f"Setting up target labels for {output_name}.")

        match output_type_info:
            case TabularOutputTypeConfig():
                test_labels_df = process_tabular_output_for_testing(
                    output_name=output_name,
                    tabular_target_files_info=tabular_target_files_info,
                    output_folder=pc.global_config.basic_experiment.output_folder,
                    ids=ids,
                    all_ids=all_ids,
                    label_transformers=label_transformers,
                    per_modality_missing_ids=per_modality_missing_ids,
                    test_labels_df=test_labels_df,
                )
            case SequenceOutputTypeConfig():
                test_labels_df = process_sequence_output_for_testing(
                    output_name=output_name,
                    ids=ids,
                    per_modality_missing_ids=per_modality_missing_ids,
                    test_labels_df=test_labels_df,
                )

            case ArrayOutputTypeConfig() | ImageOutputTypeConfig():
                test_labels_df = process_array_or_image_output_for_testing(
                    output_name=output_name,
                    ids=ids,
                    output_config=output_config,
                    output_source=output_source,
                    per_modality_missing_ids=per_modality_missing_ids,
                    test_labels_df=test_labels_df,
                )

            case SurvivalOutputTypeConfig():
                test_labels_df = process_survival_output_for_testing(
                    output_name=output_name,
                    tabular_target_files_info=tabular_target_files_info,
                    output_folder=pc.global_config.basic_experiment.output_folder,
                    ids=ids,
                    all_ids=all_ids,
                    label_transformers=label_transformers,
                    test_labels_df=test_labels_df,
                    per_modality_missing_ids=per_modality_missing_ids,
                )
            case _:
                raise NotImplementedError(
                    f"Output type {output_type_info} not implemented"
                )

    missing_target_info = get_missing_targets_info(
        missing_ids_per_modality=per_modality_missing_ids,
    )

    test_labels_object = MergedPredictTargetLabels(
        predict_labels=test_labels_df,
        label_transformers=label_transformers,
        missing_ids_per_output=missing_target_info,
    )

    return test_labels_object


def process_tabular_output_for_testing(
    output_name: str,
    tabular_target_files_info: dict[str, Any],
    output_folder: str,
    ids: Sequence[str],
    all_ids: Set[str],
    label_transformers: dict[str, Any],
    per_modality_missing_ids: dict[str, set[str]],
    test_labels_df: pl.DataFrame,
) -> pl.DataFrame:
    cur_tabular_info = tabular_target_files_info[output_name]
    cur_labels = get_tabular_target_labels_for_predict(
        output_folder=output_folder,
        tabular_info=cur_tabular_info,
        output_name=output_name,
        ids=ids,
    )
    label_transformers[output_name] = cur_labels.label_transformers

    all_labels = cur_labels.all_labels
    cur_ids = set(all_labels.get_column("ID").to_list())
    missing_ids = all_ids.difference(cur_ids)
    per_modality_missing_ids[output_name] = missing_ids

    test_labels_df = update_labels_df(
        master_df=test_labels_df,
        new_labels=cur_labels.predict_labels,
        output_name=output_name,
    )

    return test_labels_df


def process_sequence_output_for_testing(
    output_name: str,
    ids: Sequence[str],
    per_modality_missing_ids: Dict[str, Set[str]],
    test_labels_df: pl.DataFrame,
) -> pl.DataFrame:
    cur_labels = set_up_delayed_predict_target_labels(
        ids=ids,
        output_name=output_name,
    )

    cur_missing_ids = gather_torch_null_missing_ids(
        labels=cur_labels.all_labels,
        output_name=output_name,
    )

    per_modality_missing_ids[output_name] = cur_missing_ids

    test_labels_df = update_labels_df(
        master_df=test_labels_df,
        new_labels=cur_labels.predict_labels,
        output_name=output_name,
    )

    return test_labels_df


def process_array_or_image_output_for_testing(
    output_name: str,
    ids: Sequence[str],
    output_config: Any,
    output_source: str,
    per_modality_missing_ids: Dict[str, Set[str]],
    test_labels_df: pl.DataFrame,
) -> pl.DataFrame:
    cur_labels = _set_up_predict_file_target_labels(
        ids=ids,
        output_config=output_config,
    )

    cur_missing_ids = gather_torch_null_missing_ids(
        labels=cur_labels.all_labels,
        output_name=output_name,
    )
    per_modality_missing_ids[output_name] = cur_missing_ids

    is_deeplake = is_deeplake_dataset(data_source=output_source)
    col_name = f"{output_name}__{output_name}"

    polars_dtype: Type[pl.Int64] | Type[pl.Utf8]
    if is_deeplake:
        polars_dtype = pl.Int64
    else:
        polars_dtype = pl.Utf8

    test_labels_df = update_labels_df(
        master_df=test_labels_df,
        new_labels=cur_labels.predict_labels,
        output_name=output_name,
    ).with_columns([pl.col(col_name).cast(polars_dtype)])

    return test_labels_df


def process_survival_output_for_testing(
    output_name: str,
    tabular_target_files_info: Dict[str, Any],
    output_folder: str,
    ids: Sequence[str],
    all_ids: Set[str],
    label_transformers: Dict[str, Any],
    per_modality_missing_ids: Dict[str, Set[str]],
    test_labels_df: pl.DataFrame,
) -> pl.DataFrame:
    """
    Process survival output data for testing phase using Polars.

    Note: Unlike training, we only transform the time values and don't process
    event indicators, as they're not needed for prediction.
    """
    cur_tabular_info = tabular_target_files_info[output_name]

    cur_labels = get_tabular_target_labels_for_predict(
        output_folder=output_folder,
        tabular_info=cur_tabular_info,
        output_name=output_name,
        ids=ids,
    )

    msg_con = "Expected exactly one continuous column for survival data"
    assert len(cur_tabular_info.con_columns) == 1, msg_con
    msg_cat = "Expected exactly one categorical column for survival data"
    assert len(cur_tabular_info.cat_columns) == 1, msg_cat

    time_column = cur_tabular_info.con_columns[0]
    event_column = cur_tabular_info.cat_columns[0]

    df_time = pl.read_csv(
        cur_tabular_info.file_path,
        columns=["ID", time_column],
    ).with_columns([pl.col("ID").cast(pl.Utf8), pl.col(time_column).cast(pl.Float32)])

    df_time_test = df_time.filter(pl.col("ID").is_in(ids))

    df_time_test, cur_labels.predict_labels = synchronize_missing_survival_values(
        df_time=df_time_test,
        df_labels=cur_labels.predict_labels,
        time_column=time_column,
        event_column=event_column,
    )

    dur_transformer = cur_labels.label_transformers[output_name][time_column]

    cur_labels.predict_labels = cur_labels.predict_labels.with_columns(
        [
            transform_durations_with_nans(
                df=df_time_test,
                time_column=time_column,
                transformer=dur_transformer,
            ).alias(time_column)
        ]
    )

    label_transformers[output_name] = cur_labels.label_transformers

    all_labels = cur_labels.all_labels
    cur_ids = set(all_labels.get_column("ID").to_list())
    missing_ids = all_ids.difference(cur_ids)
    per_modality_missing_ids[output_name] = missing_ids

    test_labels_df = update_labels_df(
        master_df=test_labels_df,
        new_labels=cur_labels.predict_labels,
        output_name=output_name,
    )

    return test_labels_df


def _set_up_predict_file_target_labels(
    ids: Sequence[str],
    output_config: OutputConfig,
) -> PredictTargetLabels:
    output_name = output_config.output_info.output_name
    output_source = output_config.output_info.output_source
    output_inner_key = output_config.output_info.output_inner_key

    id_to_data_pointer_mapping = gather_data_pointers_from_data_source(
        data_source=Path(output_source),
        validate=True,
        output_inner_key=output_inner_key,
    )

    values = [id_to_data_pointer_mapping.get(id_, None) for id_ in ids]

    df_labels = pl.DataFrame(
        {
            "ID": list(ids),
            f"{output_name}": values,
        }
    )

    df_labels = df_labels.with_columns([pl.col(f"{output_name}").cast(pl.Utf8)])

    return PredictTargetLabels(
        predict_labels=df_labels,
        label_transformers={},
    )


def set_up_delayed_predict_target_labels(
    ids: Sequence[str],
    output_name: str,
) -> PredictTargetLabels:
    df_labels = pl.DataFrame(
        {
            "ID": list(ids),
            f"{output_name}": ["DELAYED"] * len(ids),
        },
        schema_overrides={f"{output_name}": pl.Categorical},
    )

    return PredictTargetLabels(
        predict_labels=df_labels,
        label_transformers={},
    )



from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Callable, Literal, Type, Union

import torch
from torch import nn

from eir.setup import schemas
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.train_utils.metrics import (
    LogEmptyLossProtocol,
    calculate_prediction_losses,
    log_empty_loss_once,
)

if TYPE_CHECKING:
    from eir.setup.output_setup import al_output_objects_as_dict
    from eir.setup.schema_modules.output_schemas_tabular import (
        al_cat_loss_names,
        al_con_loss_names,
    )

al_cat_losses = nn.CrossEntropyLoss
al_con_losses = (
    nn.MSELoss | nn.L1Loss | nn.SmoothL1Loss | nn.PoissonNLLLoss | nn.HuberLoss
)

al_criteria = al_con_losses | al_cat_losses | Callable

al_criteria_dict = dict[str, al_criteria]

al_losses = (
    nn.CrossEntropyLoss
    | nn.MSELoss
    | nn.L1Loss
    | nn.SmoothL1Loss
    | nn.PoissonNLLLoss
    | nn.HuberLoss
    | Callable
)


al_losses_classes = (
    Type[nn.CrossEntropyLoss]
    | Type[nn.MSELoss]
    | Type[nn.L1Loss]
    | Type[nn.SmoothL1Loss]
    | Type[nn.PoissonNLLLoss]
    | Type[nn.HuberLoss]
)


def get_criteria(outputs_as_dict: "al_output_objects_as_dict") -> al_criteria_dict:
    criteria_dict: al_criteria_dict = {}
    log_empty_once = log_empty_loss_once()

    for output_name, output_object in outputs_as_dict.items():

        match output_object:
            case ComputedTabularOutputInfo():

                con_loss_name = _parse_loss_name(
                    output_config=output_object.output_config,
                    column_type="con",
                )

                criterion_con = get_supervised_criterion(
                    column_type_="con",
                    loss_name=con_loss_name,
                    reduction="none",
                )

                label_smoothing = _get_label_smoothing(
                    output_config=output_object.output_config,
                    column_type="cat",
                )
                cat_loss_name = _get_cat_loss_name(
                    output_config=output_object.output_config
                )
                criterion_cat = get_supervised_criterion(
                    column_type_="cat",
                    loss_name=cat_loss_name,
                    cat_label_smoothing_=label_smoothing,
                    reduction="none",
                )
                cat_loss_callable = (
                    loop_ce_loss
                    if cat_loss_name == "CrossEntropyLoss"
                    else vectorized_bce_loss
                )
                assert isinstance(
                    criterion_cat, (nn.CrossEntropyLoss, nn.BCEWithLogitsLoss)
                )

                loss_func = partial(
                    tabular_output_loss,
                    con_columns=output_object.target_columns["con"],
                    cat_columns=output_object.target_columns["cat"],
                    output_name=output_name,
                    log_empty_loss_callable=log_empty_once,
                    con_loss_func=criterion_con,
                    cat_loss_func=criterion_cat,
                    cat_loss_callable=cat_loss_callable,
                )

                criteria_dict[output_name] = loss_func

            case ComputedSequenceOutputInfo():
                pad_token = getattr(output_object.tokenizer, "pad_token", "<pad>")
                pad_idx = output_object.vocab[pad_token]

                criterion = partial(
                    _sequence_cat_loss,
                    cat_loss_func=nn.CrossEntropyLoss(ignore_index=pad_idx),
                )

                criterion_callable = _get_extract_and_call_criterion(
                    output_name=output_name,
                    loss_callable=criterion,
                )

                criteria_dict[output_name] = criterion_callable

            case ComputedArrayOutputInfo() | ComputedImageOutputInfo():
                criterion = partial(
                    _calc_con_loss,
                    loss_func=nn.MSELoss(),
                    nan_handling=NaNHandling.MASK,
                )

                criterion_callable = _get_extract_and_call_criterion(
                    output_name=output_name,
                    loss_callable=criterion,
                )

                criteria_dict[output_name] = criterion_callable

            case ComputedSurvivalOutputInfo():
                output_type_info = output_object.output_config.output_type_info
                assert isinstance(output_type_info, schemas.SurvivalOutputTypeConfig)

                surv_loss_name = output_type_info.loss_function
                time_column = output_type_info.time_column

                criterion_surv = create_survival_criterion(
                    loss_name=surv_loss_name,
                    time_column=time_column,
                )
                criteria_dict[output_name] = criterion_surv

    return criteria_dict


def _get_extract_and_call_criterion(
    output_name: str,
    loss_callable: Callable,
) -> Callable:

    def _compute_loss(
        input: dict[str, torch.Tensor], target: dict[str, torch.Tensor]
    ) -> dict[str, torch.Tensor]:
        input_tensor = input[output_name]
        target_tensor = target[output_name]
        loss = loss_callable(input=input_tensor, target=target_tensor)
        return {output_name: loss}

    return _compute_loss


def tabular_output_loss(
    predictions: dict[str, torch.Tensor],
    targets: dict[str, torch.Tensor],
    con_columns: list[str],
    cat_columns: list[str],
    output_name: str,
    log_empty_loss_callable: LogEmptyLossProtocol,
    con_loss_func: Callable,
    cat_loss_func: nn.CrossEntropyLoss | nn.BCEWithLogitsLoss,
    cat_loss_callable: Callable,
) -> dict[str, torch.Tensor]:
    con_preds = {k: predictions[k] for k in con_columns}
    cat_preds = {k: predictions[k] for k in cat_columns}
    con_targets = {k: targets[k] for k in con_columns}
    cat_targets = {k: targets[k] for k in cat_columns}

    con_losses = (
        vectorized_con_loss(
            predictions=con_preds,
            targets=con_targets,
            loss_func=con_loss_func,
        )
        if con_columns
        else {}
    )

    cat_losses = (
        cat_loss_callable(
            predictions=cat_preds,
            targets=cat_targets,
            loss_func=cat_loss_func,
        )
        if cat_columns
        else {}
    )

    all_losses = {}
    for name, loss in con_losses.items():
        if torch.isnan(loss):
            log_empty_loss_callable(output_name=output_name, output_head_name=name)
        all_losses[name] = loss
    for name, loss in cat_losses.items():
        if torch.isnan(loss):
            log_empty_loss_callable(output_name=output_name, output_head_name=name)
        all_losses[name] = loss

    return all_losses


def vectorized_con_loss(
    predictions: dict[str, torch.Tensor],
    targets: dict[str, torch.Tensor],
    loss_func: Callable = nn.MSELoss(reduction="none"),
) -> dict[str, torch.Tensor]:
    predictions_stacked = torch.stack(list(predictions.values()), dim=1).squeeze()
    targets_stacked = torch.stack(list(targets.values()), dim=1).squeeze()

    valid_mask = ~torch.isnan(targets_stacked)

    predictions_masked = predictions_stacked * valid_mask
    targets_masked = targets_stacked.clone()
    targets_masked[~valid_mask] = 0

    losses = loss_func(predictions_masked, targets_masked)
    target_means = losses.sum(dim=0) / (valid_mask.sum(dim=0) + 1e-8)

    if len(target_means.shape) == 0:
        target_means = target_means.unsqueeze(0)

    return {name: loss for name, loss in zip(predictions.keys(), target_means)}


def vectorized_bce_loss(
    predictions: dict[str, torch.Tensor],
    targets: dict[str, torch.Tensor],
    loss_func: Callable = nn.BCEWithLogitsLoss(reduction="none"),
) -> dict[str, torch.Tensor]:
    predictions_stacked = torch.stack(list(predictions.values()), dim=1).squeeze()
    targets_stacked = torch.stack(list(targets.values()), dim=1).squeeze().float()

    valid_mask = targets_stacked != -1

    targets_masked = targets_stacked.clone()
    targets_masked[~valid_mask] = 0

    losses = loss_func(predictions_stacked, targets_masked)
    target_means = losses.sum(dim=0) / (valid_mask.sum(dim=0) + 1e-8)

    if len(target_means.shape) == 0:
        target_means = target_means.unsqueeze(0)

    return {name: loss for name, loss in zip(predictions.keys(), target_means)}


def loop_ce_loss(
    predictions: dict[str, torch.Tensor],
    targets: dict[str, torch.Tensor],
    loss_func: Callable = nn.CrossEntropyLoss(reduction="none"),
) -> dict[str, torch.Tensor]:
    target_losses = {}

    for name in predictions:
        valid_mask = targets[name] != -1

        pred = predictions[name]
        target = targets[name]

        masked_target = target.clone()
        masked_target[~valid_mask] = -100

        losses = loss_func(pred, masked_target)

        target_losses[name] = (
            losses[valid_mask].mean() if valid_mask.any() else losses.new_zeros(())
        )

    return target_losses


def build_loss_dict() -> dict[str, list[str]]:
    loss_dict = {
        "cat": [
            "CrossEntropyLoss",
            "BCEWithLogitsLoss",
        ],
        "con": [
            "MSELoss",
            "L1Loss",
            "SmoothL1Loss",
            "PoissonNLLLoss",
            "HuberLoss",
        ],
    }

    return loss_dict


def get_supervised_criterion(
    column_type_: str,
    loss_name: str,
    cat_label_smoothing_: float = 0.0,
    reduction: str = "none",
) -> nn.CrossEntropyLoss | Callable:
    loss_dict = build_loss_dict()

    match column_type_, loss_name:
        case "con", _:
            assert loss_name in loss_dict["con"]

            loss_module = getattr(nn, loss_name)
            return partial(
                _calc_con_loss,
                loss_func=loss_module(reduction=reduction),
            )

        case "cat", "CrossEntropyLoss":
            return nn.CrossEntropyLoss(
                label_smoothing=cat_label_smoothing_,
                reduction=reduction,
            )

        case "cat", _:
            assert loss_name in loss_dict["cat"]
            loss_module = getattr(nn, loss_name)
            return loss_module(reduction=reduction)

    raise ValueError()


def _parse_loss_name(
    output_config: schemas.OutputConfig, column_type: str
) -> Union["al_cat_loss_names", "al_con_loss_names"]:
    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)
    match column_type:
        case "cat":
            return output_type_info.cat_loss_name
        case "con":
            return output_type_info.con_loss_name
        case _:
            raise ValueError(f"Unknown column type: {column_type}")


def _get_label_smoothing(
    output_config: schemas.OutputConfig,
    column_type: str,
) -> float:
    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)
    match column_type:
        case "con":
            return 0.0
        case "cat":
            return output_type_info.cat_label_smoothing
        case _:
            raise ValueError(f"Unknown column type: {column_type}")


def _get_cat_loss_name(
    output_config: schemas.OutputConfig,
) -> "al_cat_loss_names":
    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, schemas.TabularOutputTypeConfig)
    return output_type_info.cat_loss_name


class NaNHandling(Enum):
    MASK = "mask"
    RAISE = "raise"
    NONE = "none"


def _calc_con_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    loss_func: Union[nn.MSELoss, nn.L1Loss, nn.PoissonNLLLoss],
    nan_handling: NaNHandling = NaNHandling.NONE,
) -> torch.Tensor:

    if nan_handling != NaNHandling.NONE:
        has_nan = torch.isnan(target).any()
        if has_nan and nan_handling == NaNHandling.RAISE:
            raise ValueError(
                f"NaN values found in target tensor with NaNHandling.RAISE specified. "
                f"NaN count: {torch.isnan(target).sum().item()}"
            )
        elif nan_handling == NaNHandling.MASK:
            mask = ~torch.isnan(target).squeeze()
            input_masked = input.squeeze()[mask]
            target_masked = target.squeeze()[mask]

            match loss_func:
                case nn.PoissonNLLLoss():
                    loss = loss_func(log_input=input_masked, target=target_masked)
                case _:
                    loss = loss_func(input=input_masked, target=target_masked)

            return loss

    match loss_func:
        case nn.PoissonNLLLoss():
            return loss_func(log_input=input.squeeze(), target=target.squeeze())
        case _:
            return loss_func(input=input.squeeze(), target=target.squeeze())


def get_loss_callable(
    criteria: al_criteria_dict,
) -> Callable:
    single_task_loss_func = partial(
        calculate_prediction_losses,
        criteria=criteria,
    )
    return single_task_loss_func


def _sequence_cat_loss(
    input,
    target,
    cat_loss_func: nn.CrossEntropyLoss,
) -> torch.Tensor:
    loss = cat_loss_func(input=input.transpose(2, 1), target=target)
    return loss


def _survival_loss(
    input: torch.Tensor,
    time: torch.Tensor,
    target: torch.Tensor,
    loss_func: Literal["NegativeLogLikelihood", "CoxPHLoss"] = "NegativeLogLikelihood",
) -> torch.Tensor:

    if loss_func == "NegativeLogLikelihood":
        return _negative_log_likelihood_loss(log_hazards=input, time=time, event=target)
    elif loss_func == "CoxPHLoss":
        return _cox_ph_loss(risk_scores=input, time=time, event=target)
    else:
        raise ValueError(f"Unsupported loss function: {loss_func}")


def _negative_log_likelihood_loss(
    log_hazards: torch.Tensor,
    time: torch.Tensor,
    event: torch.Tensor,
) -> torch.Tensor:
    valid_mask = (time != -1) & (event != -1)

    # Early return if no valid samples
    if not valid_mask.any():
        return log_hazards.new_zeros((), requires_grad=True)

    masked_log_hazards = log_hazards[valid_mask]
    masked_time = time[valid_mask]
    masked_event = event[valid_mask]

    # Convert log-hazards to hazards
    hazards = torch.sigmoid(masked_log_hazards)

    # Calculate log(survival probability) for each interval
    log_surv = torch.log(1 - hazards + 1e-8)

    # Calculate cumulative log(survival probability)
    cum_log_surv = torch.cumsum(log_surv, dim=1)

    # Select the relevant timepoints
    idx = torch.arange(masked_log_hazards.shape[0], device=masked_log_hazards.device)
    relevant_log_surv = cum_log_surv[idx, masked_time]
    relevant_hazards = hazards[idx, masked_time]

    # Calculate log-likelihood
    uncensored_ll = torch.log(relevant_hazards + 1e-8) + relevant_log_surv
    censored_ll = relevant_log_surv

    # Combine censored and uncensored log-likelihoods
    log_likelihood = masked_event * uncensored_ll + (1 - masked_event) * censored_ll

    # Return negative average log-likelihood
    return -torch.mean(log_likelihood)


def _cox_ph_loss(
    risk_scores: torch.Tensor,
    time: torch.Tensor,
    event: torch.Tensor,
    ties_method: str = "efron",
) -> torch.Tensor:
    valid_mask = (time != -1) & (event != -1)

    # Early returns if no valid samples
    if not valid_mask.any():
        return risk_scores.new_zeros((), requires_grad=True)

    masked_risk_scores = risk_scores[valid_mask]
    masked_time = time[valid_mask]
    masked_event = event[valid_mask]

    if torch.sum(masked_event) == 0:
        return risk_scores.new_zeros((), requires_grad=True)

    sorted_time, indices = torch.sort(masked_time)
    sorted_risk_scores = masked_risk_scores[indices]
    sorted_event = masked_event[indices]

    if ties_method == "breslow":
        log_risk = torch.logcumsumexp(sorted_risk_scores.flip(0), dim=0).flip(0)
        loss = sorted_risk_scores - log_risk
        loss = loss[sorted_event == 1]
        return -torch.mean(loss)

    elif ties_method == "efron":
        unique_times = torch.unique(sorted_time)

        log_likelihood = torch.tensor(0.0, device=risk_scores.device)

        for t in unique_times:
            events_at_t = (sorted_time == t) & (sorted_event == 1)
            if not torch.any(events_at_t):
                continue

            # Risk set (samples still at risk at time t)
            risk_set = sorted_time >= t

            # Calculate Efron's correction for ties
            n_events = events_at_t.sum()
            tied_scores = sorted_risk_scores[events_at_t]
            risk_scores_exp = torch.exp(sorted_risk_scores)
            risk_set_scores = risk_scores_exp[risk_set]
            tied_scores_exp = risk_scores_exp[events_at_t]

            for j in range(n_events):
                factor = j / n_events
                log_likelihood += tied_scores[j].squeeze() - torch.log(
                    torch.sum(risk_set_scores) - factor * torch.sum(tied_scores_exp)
                )

        n_events_total = torch.sum(masked_event)
        return -log_likelihood / n_events_total

    else:
        raise ValueError(f"Unsupported ties method: {ties_method}")


def create_survival_criterion(
    loss_name: Literal["NegativeLogLikelihood", "CoxPHLoss"],
    time_column: str,
) -> Callable:

    def survival_criterion(
        predictions: dict[str, torch.Tensor],
        targets: dict[str, torch.Tensor],
    ) -> dict[str, torch.Tensor]:
        time_tensor = targets[time_column].to(dtype=torch.long)

        losses = {}
        for name, pred in predictions.items():
            if name != time_column:
                losses[name] = _survival_loss(
                    input=pred,
                    time=time_tensor,
                    target=targets[name],
                    loss_func=loss_name,
                )
        return losses

    return survival_criterion



from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from eir.setup.config_setup_modules.input_config_initialization import init_input_config

if TYPE_CHECKING:
    from eir.setup.schemas import InputConfig


def load_input_config_from_yaml(input_config_path: Path) -> "InputConfig":
    with open(input_config_path, "r") as infile:
        input_config_dict = yaml.safe_load(stream=infile)

    input_config = init_input_config(yaml_config_as_dict=input_config_dict)

    return input_config



import json
from copy import deepcopy
from pathlib import Path

from eir.experiment_io.input_object_io_modules.input_io_utils import (
    load_input_config_from_yaml,
)
from eir.setup.input_setup_modules.setup_sequence import (
    ComputedSequenceInputInfo,
    set_up_computed_sequence_input,
)
from eir.setup.schemas import SequenceInputDataConfig


def load_sequence_input_object(
    serialized_input_folder: Path,
) -> ComputedSequenceInputInfo:
    config_path = serialized_input_folder / "input_config.yaml"
    vocab_json_path = _get_vocab_path(serialized_input_folder=serialized_input_folder)
    computed_max_length_path = serialized_input_folder / "computed_max_length.json"

    input_config = load_input_config_from_yaml(input_config_path=config_path)
    computed_max_length = json.loads(computed_max_length_path.read_text())

    input_config_modified = deepcopy(input_config)
    input_type_info_modified = deepcopy(input_config.input_type_info)
    assert isinstance(input_type_info_modified, SequenceInputDataConfig)

    input_type_info_modified.vocab_file = str(vocab_json_path)
    input_type_info_modified.max_length = computed_max_length

    input_config_modified.input_type_info = input_type_info_modified

    loaded_object = set_up_computed_sequence_input(
        input_config=input_config_modified,
        mode="eval",
    )

    return loaded_object


def _get_vocab_path(serialized_input_folder: Path) -> Path:
    vocab_json_path = serialized_input_folder / "vocab.json"
    bpe_tokenizer_path = serialized_input_folder / "bpe_tokenizer.json"
    if bpe_tokenizer_path.exists():
        return bpe_tokenizer_path
    return vocab_json_path



from typing import Any, Literal, Optional

import torch
from einops import rearrange
from torch import Tensor, einsum, nn

from eir.models.layers.attention_layers import SwiGLU, Transformer
from eir.models.layers.lcl_layers import LCL, LCLResidualBlock
from eir.models.layers.projection_layers import get_1d_projection_layer

al_projection_layer_types = Literal["auto", "lcl", "lcl_residual", "linear"]


class MetaSequenceProjection(nn.Module):
    def __init__(
        self,
        in_total_num_elements: int,
        in_embedding_dim: int,
        target_embedding_dim: int,
        target_max_length: int,
        projection_layer_type: al_projection_layer_types = "auto",
        *args,
        **kwargs,
    ):
        super().__init__()

        self.in_total_num_elements = in_total_num_elements
        self.in_embedding_dim = in_embedding_dim
        self.target_embedding_dim = target_embedding_dim
        self.target_max_length = target_max_length

        self.projection = SequenceProjection(
            in_features=in_total_num_elements,
            target_embedding_dim=self.target_embedding_dim,
            target_max_length=self.target_max_length,
            projection_layer_type=projection_layer_type,
        )

        self.cross_attention_projection = SequenceResidualCrossAttentionProjection(
            in_embedding_dim=self.in_embedding_dim,
            target_embedding_dim=self.target_embedding_dim,
            target_max_length=self.target_max_length,
        )

        self.scaling_factors = nn.Parameter(torch.ones(3))

    def forward(
        self, input_tensor: torch.Tensor, target_tensor: torch.Tensor
    ) -> torch.Tensor:
        weights = torch.softmax(self.scaling_factors, dim=0)

        projected = self.projection(input_tensor)
        projected = projected * weights[0]

        cross_attended = self.cross_attention_projection(
            x=target_tensor, context=input_tensor
        )
        cross_attended = cross_attended * weights[1]

        identity = target_tensor * weights[2]

        out = projected + cross_attended + identity

        return out


class SequenceProjection(nn.Module):
    def __init__(
        self,
        in_features: int,
        target_embedding_dim: int,
        target_max_length: int,
        projection_layer_type: al_projection_layer_types = "auto",
        *args,
        **kwargs,
    ):
        super().__init__()

        self.in_features = in_features
        self.target_embedding_dim = target_embedding_dim
        self.target_max_length = target_max_length
        self.projection_layer_type = projection_layer_type

        self.out_dim = self.target_max_length * self.target_embedding_dim

        self.norm_1 = nn.RMSNorm(normalized_shape=in_features)
        self.projection_layer = get_1d_projection_layer(
            input_dimension=in_features,
            target_dimension=self.out_dim,
            projection_layer_type=self.projection_layer_type,
        )

        self.act = SwiGLU(
            in_features=target_embedding_dim,
            hidden_features=target_embedding_dim * 4,
            out_features=target_embedding_dim,
            bias=False,
        )

        self.encoder = Transformer(
            d_model=target_embedding_dim,
            nhead=8,
            num_layers=1,
            dim_feedforward=target_embedding_dim * 4,
            dropout=0.1,
            norm_first=True,
        )

        self.downsample_identity: nn.Linear | nn.Identity | LCL | LCLResidualBlock = (
            nn.Identity()
        )
        if self.in_features != self.out_dim:
            self.downsample_identity = get_1d_projection_layer(
                input_dimension=in_features,
                target_dimension=self.out_dim,
                projection_layer_type=self.projection_layer_type,
            )

    def forward(self, input: Tensor) -> Tensor:
        input_flat = input.flatten(1)

        identity = self.downsample_identity(input_flat)[..., : self.out_dim]
        identity = identity.reshape(
            identity.shape[0],
            self.target_max_length,
            self.target_embedding_dim,
        )

        out = self.norm_1(input_flat)
        out = self.projection_layer(out)[..., : self.out_dim]

        out = out.reshape(
            out.shape[0],
            self.target_max_length,
            self.target_embedding_dim,
        )
        out = self.act(out)

        out = self.encoder(out)

        return out + identity


class SequenceResidualCrossAttentionProjection(nn.Module):
    def __init__(
        self,
        in_embedding_dim: int,
        target_embedding_dim: int,
        target_max_length: int,
        *args,
        **kwargs,
    ):
        super().__init__()

        self.in_embedding_dim = in_embedding_dim
        self.target_embedding_dim = target_embedding_dim
        self.target_max_length = target_max_length

        self.projection_layer = UniDirectionalCrossAttention(
            dim=self.target_embedding_dim,
            context_dim=self.in_embedding_dim,
            dim_head=self.target_embedding_dim,
            dropout=0.1,
            talking_heads=False,
            pre_norm=False,
        )

        self.norm_1_target = nn.RMSNorm(normalized_shape=target_embedding_dim)
        self.act_1 = SwiGLU(
            in_features=target_embedding_dim,
            hidden_features=target_embedding_dim,
            out_features=target_embedding_dim,
            bias=False,
        )

        self.norm_1_context = nn.RMSNorm(normalized_shape=in_embedding_dim)
        self.act_context = SwiGLU(
            in_features=in_embedding_dim,
            hidden_features=in_embedding_dim,
            out_features=in_embedding_dim,
            bias=False,
        )

        self.encoder = Transformer(
            d_model=target_embedding_dim,
            nhead=8,
            num_layers=1,
            dim_feedforward=target_embedding_dim * 4,
            dropout=0.1,
            norm_first=True,
        )

        ca_mask = torch.ones((1, self.target_max_length)).bool()
        self.register_buffer("ca_mask", ca_mask)

        encoder_mask = torch.triu(
            torch.ones(self.target_max_length, self.target_max_length) * float("-inf"),
            diagonal=1,
        )
        self.register_buffer("encoder_mask", encoder_mask)

        self.norm_2_target = nn.RMSNorm(normalized_shape=target_embedding_dim)
        self.act_2 = SwiGLU(
            in_features=target_embedding_dim,
            hidden_features=target_embedding_dim,
            out_features=target_embedding_dim,
            bias=False,
        )

        self.downsample_identity = UniDirectionalCrossAttention(
            dim=self.target_embedding_dim,
            context_dim=self.in_embedding_dim,
            dim_head=self.target_embedding_dim,
            dropout=0.1,
            talking_heads=False,
            pre_norm=False,
        )

    def forward(self, x: torch.Tensor, context: torch.Tensor) -> torch.Tensor:
        context = context.reshape(context.shape[0], -1, self.in_embedding_dim)

        identity = self.downsample_identity(x=x, context=context, mask=self.ca_mask)

        out = self.norm_1_target(x)
        out = self.act_1(out)

        out_context = self.norm_1_context(context)
        out_context = self.act_context(out_context)

        out = self.projection_layer(x=out, context=out_context, mask=self.ca_mask)

        out = self.norm_2_target(out)
        out = self.act_2(out)
        out = self.encoder(out, mask=self.encoder_mask)

        return out + identity


def exists(val: Any) -> bool:
    return val is not None


def default(val: Any, d: Any) -> Any:
    return val if exists(val) else d


def stable_softmax(t: torch.Tensor, dim: int = -1) -> torch.Tensor:
    t = t - t.amax(dim=dim, keepdim=True)
    return t.softmax(dim=dim)


class UniDirectionalCrossAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        context_dim: int,
        heads: int = 8,
        dim_head: int = 64,
        dropout: float = 0.0,
        talking_heads: bool = False,
        pre_norm: bool = False,
    ):
        """
        Adapted from: https://github.com/lucidrains/bidirectional-cross-attention
        """
        super().__init__()
        context_dim = default(val=context_dim, d=dim)

        self.norm = nn.RMSNorm(normalized_shape=dim) if pre_norm else nn.Identity()
        self.context_norm = (
            nn.RMSNorm(normalized_shape=context_dim) if pre_norm else nn.Identity()
        )

        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = dim_head * heads

        self.dropout = nn.Dropout(p=dropout)

        self.to_qk = nn.Linear(
            in_features=dim,
            out_features=inner_dim,
            bias=False,
        )
        self.context_to_qk = nn.Linear(
            in_features=context_dim,
            out_features=inner_dim,
            bias=False,
        )

        self.to_v = nn.Linear(
            in_features=dim,
            out_features=inner_dim,
            bias=False,
        )
        self.context_to_v = nn.Linear(
            in_features=context_dim,
            out_features=inner_dim,
            bias=False,
        )

        self.to_out = nn.Linear(
            in_features=inner_dim,
            out_features=dim,
        )

        self.talking_heads = (
            nn.Conv2d(
                in_channels=heads,
                out_channels=heads,
                kernel_size=1,
                bias=False,
            )
            if talking_heads
            else nn.Identity()
        )

    def forward(
        self,
        x: torch.Tensor,
        context: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        context_mask: Optional[torch.Tensor] = None,
        rel_pos_bias: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        _, i, j, h, device = (
            x.shape[0],
            x.shape[-2],
            context.shape[-2],
            self.heads,
            x.device,
        )

        x = self.norm(x)
        context = self.context_norm(context)

        # get shared query/keys and values for sequence and context
        qk, v = self.to_qk(x), self.to_v(x)
        context_qk, context_v = self.context_to_qk(context), self.context_to_v(context)

        # split out head
        qk, context_qk, v, context_v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=h),
            (qk, context_qk, v, context_v),
        )

        # get similarities
        sim = einsum("b h i d, b h j d -> b h i j", qk, context_qk) * self.scale

        # relative positional bias, if supplied
        if exists(val=rel_pos_bias):
            sim = sim + rel_pos_bias

        # causal mask
        if exists(val=mask) or exists(val=context_mask):
            i, j = sim.shape[-2:]
            mask = torch.ones(i, j, device=device, dtype=torch.bool).triu(j - i + 1)
            mask_value = -torch.finfo(sim.dtype).max
            sim = sim.masked_fill(mask, mask_value)

        # get attention along sequence length, using shared similarity matrix
        attn = stable_softmax(t=sim, dim=-1)

        attn = self.dropout(attn)

        # talking heads
        attn = self.talking_heads(attn)

        # src sequence aggregates values from context
        out = einsum("b h i j, b h j d -> b h i d", attn, context_v)

        # merge heads and combine out
        out = rearrange(out, "b h n d -> b n (h d)")

        out = self.to_out(out)

        return out



from collections import Counter, OrderedDict
from typing import Iterable, Optional

from eir.setup.input_setup_modules.torchtext_port.vocab import Vocab


def vocab(
    ordered_dict: dict[str, int],
    min_freq: int = 1,
    specials: Optional[list[str]] = None,
    special_first: bool = True,
) -> Vocab:
    """
    Factory method for creating a vocab object which maps tokens to indices.

    Note that the ordering in which key-value pairs were inserted in the
    `ordered_dict` will be respected when building the vocab.
    Therefore, if sorting by token frequency is important to the user, the
    `ordered_dict` should be created in a way to reflect this.

    Args:
        ordered_dict: Ordered Dictionary mapping tokens to their
        corresponding occurrence frequencies.
        min_freq: The minimum frequency needed to include a token in the vocabulary.
        specials: Special symbols to add. The order of supplied tokens will be
        preserved.
        special_first: Indicates whether to insert special symbols at the beginning or
        at the end.

    Returns:
        Vocab: A `Vocab` object

    Examples:
        >>> ordered_dict = OrderedDict([('a', 5), ('b', 3), ('c', 1)])
        >>> v1 = vocab(ordered_dict)
        >>> print(v1['a'])  # prints 0
        >>> print(v1['c'])  # prints 2
        >>> v2 = vocab(ordered_dict, specials=['<unk>'])
        >>> print(v2['<unk>'])  # prints 0
        >>> print(v2['a'])  # prints 1
        >>> v2.set_default_index(v2['<unk>'])
        >>> print(v2['out of vocab'])  # prints 0 (index of <unk>)
    """
    specials = specials or []
    tokens = []

    # Add special tokens
    if special_first:
        tokens.extend(specials)

    # Add tokens that meet the minimum frequency
    for token, freq in ordered_dict.items():
        if freq >= min_freq and token not in specials:
            tokens.append(token)

    # Add special tokens at the end if not added at the beginning
    if not special_first:
        tokens.extend(specials)

    # Create and return the Vocab object
    return Vocab(tokens)


def build_vocab_from_iterator(
    iterator: Iterable,
    min_freq: int = 1,
    specials: Optional[list[str]] = None,
    special_first: bool = True,
    max_tokens: Optional[int] = None,
    sort_by_freq: bool = False,
) -> Vocab:
    """
    Build a Vocab from an iterator.

    Args:
        iterator: Iterator used to build Vocab. Must yield list or iterator of tokens.
        min_freq: The minimum frequency needed to include a token in the vocabulary.
        specials: Special symbols to add. The order of supplied tokens will be
        preserved.
        special_first: Indicates whether to insert special symbols at the beginning
        or at the end.
        max_tokens: If provided, creates the vocab from the
        `max_tokens - len(specials)` most frequent tokens.
        sort_by_freq: If true, sort the vocab by frequency in descending order.

    Returns:
        Vocab: A `Vocab` object

    Examples:
        >>> def yield_tokens(texts):
        ...     for text in texts:
        ...         yield text.split()
        >>> texts = ["the quick brown fox", "jumps over the lazy dog", ... ]
        >>> vocab = build_vocab_from_iterator(yield_tokens(texts), specials=["<unk>"])
        >>> print(vocab.get_itos())  # prints ['<unk>', 'the', 'lazy', 'dog', ...]
        >>> vocab.set_default_index(vocab['<unk>'])
        >>> print(vocab['river'])  # prints 0 (index of <unk>)

    Note:
        We have the isinstance check there as when e.g. reading from a vocabulary file,
        each token will be yielded as a string, not as a list of strings (as hen
        e.g. yielding from files split on e.g. whitespace). When the Counter just
        gets a string, it will count the individual characters, which is not what we
        want.

        We have the sort_by_freq parameter to ensure that e.g. when building a vocab
        from a vocab file, the order of the tokens in the file is preserved.

    """
    counter: Counter = Counter()
    for tokens in iterator:
        if isinstance(tokens, str):
            tokens = [tokens]
        counter.update(tokens)

    specials = specials or []

    if sort_by_freq:
        # Sort by descending frequency, then lexicographically
        sorted_order = sorted(counter.items(), key=lambda x: (-x[1], x[0]))
    else:
        sorted_order = list(counter.items())

    if max_tokens is not None:
        if len(specials) >= max_tokens:
            raise ValueError(
                "len(specials) >= max_tokens, so the "
                "vocab will be entirely special tokens."
            )
        sorted_order = sorted_order[: max_tokens - len(specials)]

    ordered_dict = OrderedDict(sorted_order)

    return vocab(
        ordered_dict=ordered_dict,
        min_freq=min_freq,
        specials=specials,
        special_first=special_first,
    )



from copy import deepcopy
from pathlib import Path
from typing import Sequence

import aislib.misc_utils
import yaml

from eir.experiment_io.io_utils import strip_config
from eir.models.input.sequence.transformer_models import (
    BasicTransformerFeatureExtractorModelConfig,
)
from eir.setup.config import (
    Configs,
    get_global_config,
    get_input_configs,
    load_fusion_configs,
    load_output_configs,
)
from eir.setup.config_setup_modules.config_setup_utils import object_to_primitives
from eir.setup.schema_modules.output_schemas_sequence import (
    SequenceOutputSamplingConfig,
)
from eir.setup.schemas import InputConfig, OutputConfig


def load_configs(configs_root_folder: Path) -> Configs:
    """
    Note we do not add the extra sequence output configs here as they
    are already added to the configs when they are saved.
    """
    global_path = configs_root_folder / "global_config.yaml"
    input_path = configs_root_folder / "input_configs.yaml"
    fusion_path = configs_root_folder / "fusion_config.yaml"
    output_path = configs_root_folder / "output_configs.yaml"

    with open(global_path, "r") as infile:
        global_config_list = [yaml.safe_load(stream=infile)]
    global_config = get_global_config(global_configs=global_config_list)

    with open(input_path, "r") as infile:
        input_config_list = yaml.safe_load(stream=infile)

    input_configs = get_input_configs(input_configs=input_config_list)
    input_configs_patched = _patch_sequence_input_configs_linked_from_output(
        input_configs=input_configs
    )

    with open(fusion_path, "r") as infile:
        fusion_config_list = [yaml.safe_load(stream=infile)]
    fusion_config = load_fusion_configs(fusion_configs=fusion_config_list)

    with open(output_path, "r") as infile:
        output_config_list = yaml.safe_load(stream=infile)
    output_configs = load_output_configs(output_configs=output_config_list)
    output_configs_patched = _patch_sequence_output_configs(
        output_configs=output_configs
    )

    aggregate_config = Configs(
        global_config=global_config,
        input_configs=input_configs_patched,
        fusion_config=fusion_config,
        output_configs=output_configs_patched,
    )

    return aggregate_config


def _patch_sequence_input_configs_linked_from_output(
    input_configs: Sequence[InputConfig],
) -> Sequence[InputConfig]:
    """
    We need this as during training when configs are set up, we call
    get_configs_object_with_seq_output_configs(configs=aggregate_config) to build
    relevant sequence input configs from the sequence output configs.

    A part of the logic there initializes the model_init_config of the input config
    to be a BasicTransformerFeatureExtractorModelConfig object.

    However, we do not reuse the get_configs_object_with_seq_output_configs function
    here as the configs have already been built (during training) and saved,
    hence we are only loading them here. Calling that function results in a
    collision / error. However, we still need to patch the model_init_config
    here to ensure the proper BasicTransformerFeatureExtractorModelConfig object
    is initialized.
    """
    parsed_configs = []

    for input_config in input_configs:
        input_type = input_config.input_info.input_type
        if input_type != "sequence":
            parsed_configs.append(input_config)
            continue

        model_type = input_config.model_config.model_type
        if not model_type.startswith("eir-input-sequence-from-linked-output-"):
            parsed_configs.append(input_config)
            continue

        input_config_copy = deepcopy(input_config)

        init_kwargs_as_dict = input_config.model_config.model_init_config
        assert isinstance(init_kwargs_as_dict, dict)

        model_init_config = BasicTransformerFeatureExtractorModelConfig(
            **init_kwargs_as_dict
        )

        input_config_copy.model_config.model_init_config = model_init_config

        parsed_configs.append(input_config_copy)

    return parsed_configs


def _patch_sequence_output_configs(
    output_configs: Sequence[OutputConfig],
) -> Sequence[OutputConfig]:
    parsed_configs = []

    for output_config in output_configs:
        output_type = output_config.output_info.output_type
        if output_type != "sequence":
            parsed_configs.append(output_config)
            continue

        init_kwargs = output_config.sampling_config
        if init_kwargs is None:
            continue

        output_config_copy = deepcopy(output_config)
        init_kwargs_as_dict = output_config.sampling_config
        assert isinstance(init_kwargs_as_dict, dict)

        sampling_config = SequenceOutputSamplingConfig(**init_kwargs_as_dict)

        output_config_copy.sampling_config = sampling_config

        parsed_configs.append(output_config_copy)

    return parsed_configs


def save_yaml_configs(
    run_folder: Path,
    configs: "Configs",
    for_serialized: bool = False,
) -> None:

    output_root = run_folder / "configs"
    if for_serialized:
        output_root = run_folder / "serializations" / "configs_stripped"

    for config_name, config_object in configs.__dict__.items():
        if config_name == "gc":
            continue

        cur_output_path = Path(output_root / config_name)
        cur_output_path = cur_output_path.with_suffix(".yaml")
        aislib.misc_utils.ensure_path_exists(path=cur_output_path, is_folder=False)

        if config_name == "global_config":
            main_keys = [
                "basic_experiment",
                "model",
                "optimization",
                "lr_schedule",
                "training_control",
                "evaluation_checkpoint",
                "attribution_analysis",
                "metrics",
                "visualization_logging",
                "latent_sampling",
            ]
            config_dict = {k: getattr(config_object, k) for k in main_keys}
            config_object_as_primitives = object_to_primitives(obj=config_dict)
        else:
            config_object_as_primitives = object_to_primitives(obj=config_object)

        if for_serialized:
            config_object_as_primitives = strip_config(
                config=config_object_as_primitives
            )

        with open(str(cur_output_path), "w") as yaml_file_handle:
            yaml.dump(data=config_object_as_primitives, stream=yaml_file_handle)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Literal, Optional, Sequence, Union

if TYPE_CHECKING:
    from eir.setup.schemas import al_max_sequence_length, al_tokenizer_choices

al_sequence_operations = Literal["autoregressive", "mlm"]


@dataclass
class SequenceOutputTypeConfig:
    """
    :param vocab_file:
        An optional text file containing pre-defined vocabulary to use
        for the training. If this is not passed in, the framework will automatically
        build the vocabulary from the training data. Passing in a vocabulary file is
        therefore useful if (a) you want to manually specify / limit the vocabulary used
        and/or (b) you want to save time by pre-computing the vocabulary. The
        expected format is a json file with an array of tokens, e.g. {"<bos>": 0,
        "<unk>": 1, "<mask>": 2, "<pad>": 3, "<eos>": 4, "a": 5, ...}.

    :param max_length:
        Maximum length to truncate/pad sequences to. This can be an integer or the
        values 'max' or 'average'. The 'max' keyword will use the maximum sequence
        length found in the training data, while the 'average' will use the average
        length across all training samples.

    :param sampling_strategy_if_longer:
        Controls how sequences are truncated if they are longer than the specified
        ``max_length`` parameter. Using 'from_start' will always truncate from the
        beginning of the sequence, ensuring the the samples will always be the same
        during training. Setting this parameter to ``uniform`` will uniformly sample
        a slice of a given sample sequence during training. Note that for consistency,
        the validation/test set samples always use the ``from_start`` setting when
        truncating.

    :param min_freq:
        Minimum number of times a token must appear in the total training data to be
        included in the vocabulary. Note that this setting will not do anything if
        passing in ``vocab_file``.

    :param split_on:
        Which token to split the sequence on to generate separate tokens for the
        vocabulary.

    :param tokenizer:
        Which tokenizer to use. Relevant if modelling on language, but not as much when
        doing it on other arbitrary sequences.

    :param tokenizer_language:
        Which language rules the tokenizer should apply when tokenizing the raw data.

    :param adaptive_tokenizer_max_vocab_size:
        If using an adaptive tokenizer ("bpe"), this parameter controls the maximum
        size of the vocabulary.

    :param sequence_operation:
        Which operation to perform on the sequence. Currently only ``autoregressive``
        is supported, which means that the model will be trained to predict the next
        token in the sequence given the previous tokens.
    """

    vocab_file: Union[None, str] = None
    max_length: "al_max_sequence_length" = "average"
    sampling_strategy_if_longer: Literal["from_start", "uniform"] = "uniform"
    min_freq: int = 10
    split_on: Optional[str] = " "
    tokenizer: "al_tokenizer_choices" = None  # type: ignore
    tokenizer_language: Union[str, None] = None
    adaptive_tokenizer_max_vocab_size: Optional[int] = None

    sequence_operation: al_sequence_operations = "autoregressive"


@dataclass
class SequenceOutputSamplingConfig:
    """
    :param manual_inputs:
        Manually specified inputs to use for sequence generation. This is useful
        if you want to generate sequences based on a specific input. Depending
        on the input type, different formats are expected:

        - ``sequence``: A string written directly in the ``.yaml`` file.
        - ``omics``: A file path to NumPy array of shape ``(4, n_SNPs)`` on disk.
        - ``image``: An image file path on disk.
        - ``tabular``: A mapping of :literal:`(column key: value)` written directly
          in the ``.yaml`` file.
        - ``array``: A file path to NumPy array on disk.
        - ``bytes``: A file path to a file on disk.


    :param n_eval_inputs:
        The number of inputs automatically sampled from the validation set for
        sequence generation.

    :param generated_sequence_length:
        The length of the output sequences that are generated.

    :param top_k:
        The number of top candidates to consider when sampling the next token
        in an output sequence. By default, the model considers the top 20 candidates

    :param top_p:
        The cumulative probability of the top candidates to consider when sampling
        the next token in an output sequence. For example, if top_p is 0.9, the model
        will stop sampling candidates once the cumulative probability of the most
        likely candidates reaches 0.9.
    """

    manual_inputs: Sequence[Dict[str, str]] = tuple()
    n_eval_inputs: int = 10

    generated_sequence_length: int = 64
    top_k: int = 20
    top_p: float = 0.9



from copy import copy
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from random import sample
from typing import TYPE_CHECKING, Literal, Sequence, Union

import polars as pl
from aislib.misc_utils import ensure_path_exists
from torch import nn
from torch.utils.data import DataLoader

from eir.data_load import label_setup
from eir.experiment_io.experiment_io import (
    LoadedTrainExperiment,
    load_serialized_train_experiment,
)
from eir.interpretation.interpretation import tabular_attribution_analysis_wrapper
from eir.predict_modules.predict_data import set_up_default_dataset
from eir.predict_modules.predict_input_setup import set_up_inputs_for_predict
from eir.predict_modules.predict_target_setup import (
    MergedPredictTargetLabels,
    get_target_labels_for_testing,
)
from eir.setup.config import Configs
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.output_setup import al_output_objects_as_dict
from eir.train import check_dataset_and_batch_size_compatibility
from eir.train_utils.step_logic import Hooks
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.predict import PredictExperiment

logger = get_logger(name=__name__)


def compute_predict_attributions(
    run_folder: Path,
    loaded_train_experiment: "LoadedTrainExperiment",
    predict_config: "PredictExperiment",
) -> None:
    """
    The reason why we have the different sources to load the configs from is that
    when the source is 'train', we need to have the full configs to set up to access
    the various paths used for training (e.g. input_source), which has been stripped
    from the serialized configs. In the case of loading from 'predict', we use
    the already set up paths / objects, e.g. selecting the predict output objects,
    overloading the train experiment with the predict model, inputs, etc.
    """
    gc = predict_config.configs.global_config

    background_source = (
        predict_config.predict_specific_cl_args.attribution_background_source
    )

    configs_source: Literal["configs", "configs_stripped"] = "configs"
    if background_source == "predict":
        configs_source = "configs_stripped"

    loaded_train_experiment = load_serialized_train_experiment(
        run_folder=run_folder,
        device=gc.be.device,
        source_folder=configs_source,
    )

    if configs_source == "configs":
        output_objects = loaded_train_experiment.outputs
    elif configs_source == "configs_stripped":
        output_objects = predict_config.outputs
    else:
        raise ValueError(
            f"Invalid configs source '{configs_source}'. "
            f"Expected 'configs' or 'configs_stripped'."
        )

    background_source_config = get_background_source_config(
        background_source_in_predict_cl_args=background_source,
        train_configs=loaded_train_experiment.configs,
        predict_configs=predict_config.configs,
    )
    background_dataloader = _get_predict_background_loader(
        batch_size=gc.be.batch_size,
        num_attribution_background_samples=gc.aa.attribution_background_samples,
        outputs_as_dict=output_objects,
        configs=background_source_config,
        dataloader_workers=gc.be.dataloader_workers,
        loaded_hooks=loaded_train_experiment.hooks,
    )

    overloaded_train_experiment = _overload_train_experiment_for_predict_attributions(
        train_config=loaded_train_experiment,
        predict_config=predict_config,
    )

    attribution_output_folder_callable = partial(
        _get_predict_attribution_output_folder_target,
        predict_output_folder=Path(
            predict_config.predict_specific_cl_args.output_folder
        ),
    )

    tabular_attribution_analysis_wrapper(
        model=predict_config.model,
        experiment=overloaded_train_experiment,
        output_folder_target_callable=attribution_output_folder_callable,
        dataset_to_interpret=predict_config.test_dataset,
        background_loader=background_dataloader,
    )


def get_background_source_config(
    background_source_in_predict_cl_args: Literal["train", "predict"],
    train_configs: Configs,
    predict_configs: Configs,
) -> Configs:
    """
    TODO:   In the case of predict, make sure background and samples analysed are
            separated.
    """
    if background_source_in_predict_cl_args == "predict":
        logger.info(
            "Background for attribution analysis will be loaded from sources "
            "passed to predict.py."
        )
        return predict_configs

    elif background_source_in_predict_cl_args == "train":
        logger.info(
            "Background for attribution analysis will be loaded from sources "
            "previously used for training run with name '%s'.",
            train_configs.gc.be.output_folder,
        )
        return train_configs

    raise ValueError("Invalid background source. Expected 'train' or 'predict'.")


@dataclass
class LoadedTrainExperimentMixedWithPredict(LoadedTrainExperiment):
    model: nn.Module
    inputs: al_input_objects_as_dict


def _overload_train_experiment_for_predict_attributions(
    train_config: LoadedTrainExperiment,
    predict_config: "PredictExperiment",
) -> "LoadedTrainExperimentMixedWithPredict":
    """
    TODO:   Possibly set inputs=None as a field in LoadedTrainExperiment that then gets
            filled with test_inputs. When we do not need the weird monkey-patching here
            of the batch_prep_hooks, as the LoadedTrainExperiment will have the
            inputs attribute.
    """
    train_experiment_copy = copy(train_config)

    mixed_experiment_kwargs = train_experiment_copy.__dict__
    mixed_experiment_kwargs["model"] = predict_config.model
    mixed_experiment_kwargs["configs"] = predict_config.configs
    mixed_experiment_kwargs["inputs"] = predict_config.inputs

    mixed_experiment = LoadedTrainExperimentMixedWithPredict(**mixed_experiment_kwargs)

    return mixed_experiment


def _get_predict_background_loader(
    batch_size: int,
    num_attribution_background_samples: int,
    dataloader_workers: int,
    configs: Configs,
    outputs_as_dict: al_output_objects_as_dict,
    loaded_hooks: Union["Hooks", None],
) -> DataLoader:
    """
    Note we have the target filtering there to ensure that the final
    dataset size matches that of sampled IDs. If not, there is a filtering step
    applied to the tabular inputs that come from set_up_inputs_from_predict,
    where the input hook `setup_tabular_input_for_testing` is called and filters
    on the IDs.

    As part of the basic processing of the tabular inputs, missing values have
    already been imputed. But, as setup_tabular_input_for_testing applies filtering,
    whereas other input loading funcs might not, we end up with the case where
    the joining of the tabular inputs and other modalities (done in the
    dataset setup) results in multiple NaNs for the tabular inputs,
    """

    background_ids_pool = label_setup.gather_all_ids_from_all_inputs(
        input_configs=configs.input_configs
    )

    target_labels = get_target_labels_for_testing(
        configs_overloaded_for_predict=configs,
        ids=background_ids_pool,
    )

    background_ids_sampled = _get_background_ids_sample(
        target_labels=target_labels,
        ids_inputs=background_ids_pool,
        num_attribution_background_samples=num_attribution_background_samples,
    )

    background_inputs_as_dict = set_up_inputs_for_predict(
        test_inputs_configs=configs.input_configs,
        ids=background_ids_sampled,
        hooks=loaded_hooks,
        output_folder=configs.gc.be.output_folder,
    )

    df_target = target_labels.predict_labels
    df_target_filtered = df_target.filter(pl.col("ID").is_in(background_ids_sampled))

    background_dataset = set_up_default_dataset(
        configs=configs,
        outputs_as_dict=outputs_as_dict,
        target_labels_df=df_target_filtered,
        inputs_as_dict=background_inputs_as_dict,
        missing_ids_per_output=target_labels.missing_ids_per_output,
    )

    assert len(background_dataset) <= num_attribution_background_samples

    check_dataset_and_batch_size_compatibility(
        dataset=background_dataset,
        batch_size=batch_size,
        name="Test attribution background",
    )
    background_loader = DataLoader(
        dataset=background_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=dataloader_workers,
    )

    return background_loader


def _get_predict_attribution_output_folder_target(
    predict_output_folder: Path, output_name: str, column_name: str, input_name: str
) -> Path:
    attribution_output_folder = (
        predict_output_folder / output_name / column_name / "attributions" / input_name
    )
    ensure_path_exists(path=attribution_output_folder, is_folder=True)

    return attribution_output_folder


def _get_background_ids_sample(
    target_labels: MergedPredictTargetLabels,
    ids_inputs: Sequence[str],
    num_attribution_background_samples: int,
) -> list[str]:
    ids_target_set = set(target_labels.predict_labels["ID"])
    ids_inputs_set = set(ids_inputs)

    ids_common: list[str] = list(ids_target_set.intersection(ids_inputs_set))

    background_ids_sampled: list[str] = sample(
        population=ids_common,
        k=num_attribution_background_samples,
    )

    return background_ids_sampled



from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from eir.setup.config_setup_modules.output_config_initialization import (
    init_output_config,
)

if TYPE_CHECKING:
    from eir.setup.schemas import OutputConfig


def load_output_config_from_yaml(output_config_path: Path) -> "OutputConfig":
    with open(output_config_path, "r") as infile:
        output_config_dict = yaml.safe_load(stream=infile)

    output_config = init_output_config(yaml_config_as_dict=output_config_dict)

    return output_config
